# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json
from datetime import datetime

import boto3

from airflow.providers.amazon.aws.operators.s3 import (
    S3CreateBucketOperator,
    S3CreateObjectOperator,
    S3DeleteBucketOperator,
)
from airflow.providers.amazon.aws.operators.sagemaker import (
    SageMakerDeleteModelOperator,
    SageMakerEndpointConfigOperator,
    SageMakerEndpointOperator,
    SageMakerModelOperator,
    SageMakerTrainingOperator,
)
from airflow.providers.amazon.aws.sensors.sagemaker import SageMakerEndpointSensor

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder, prune_logs

DAG_ID = "example_sagemaker_endpoint"

# Externally fetched variables:
ROLE_ARN_KEY = "ROLE_ARN"

sys_test_context_task = SystemTestContextBuilder().add_variable(ROLE_ARN_KEY).build()

# The URI of a Docker image for handling KNN model training.
# To find the URI of a free Amazon-provided image that can be used, substitute your
# desired region in the following link and find the URI under "Registry Path".
# https://docs.aws.amazon.com/sagemaker/latest/dg/ecr-us-east-1.html#knn-us-east-1.title
# This URI should be in the format of {12-digits}.dkr.ecr.{region}.amazonaws.com/knn
KNN_IMAGES_BY_REGION = {
    "us-east-1": "382416733822.dkr.ecr.us-east-1.amazonaws.com/knn:1",
    "us-west-2": "174872318107.dkr.ecr.us-west-2.amazonaws.com/knn:1",
}

# For an example of how to obtain the following train and test data, please see
# https://github.com/apache/airflow/blob/main/providers/amazon/tests/system/amazon/aws/example_sagemaker.py
TRAIN_DATA = "0,4.9,2.5,4.5,1.7\n1,7.0,3.2,4.7,1.4\n0,7.3,2.9,6.3,1.8\n2,5.1,3.5,1.4,0.2\n"
SAMPLE_TEST_DATA = "6.4,3.2,4.5,1.5"


@task
def call_endpoint(endpoint_name):
    response = (
        boto3.Session()
        .client("sagemaker-runtime")
        .invoke_endpoint(
            EndpointName=endpoint_name,
            ContentType="text/csv",
            Body=SAMPLE_TEST_DATA,
        )
    )

    return json.loads(response["Body"].read().decode())["predictions"]


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_endpoint_config(endpoint_config_job_name):
    boto3.client("sagemaker").delete_endpoint_config(EndpointConfigName=endpoint_config_job_name)


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_endpoint(endpoint_name):
    boto3.client("sagemaker").delete_endpoint(EndpointName=endpoint_name)


@task(trigger_rule=TriggerRule.ALL_DONE)
def archive_logs(log_group_name):
    boto3.client("logs").put_retention_policy(logGroupName=log_group_name, retentionInDays=1)


@task
def set_up(env_id, role_arn, ti=None):
    bucket_name = f"{env_id}-sagemaker"
    input_data_s3_key = f"{env_id}/input-data"
    training_output_s3_key = f"{env_id}/results"

    endpoint_config_job_name = f"{env_id}-endpoint-config"
    endpoint_name = f"{env_id}-endpoint"
    model_name = f"{env_id}-KNN-model"
    training_job_name = f"{env_id}-train"

    region = boto3.session.Session().region_name
    try:
        knn_image_uri = KNN_IMAGES_BY_REGION[region]
    except KeyError:
        raise KeyError(
            f"Region name {region} does not have a known KNN "
            f"Image URI.  Please add the region and URI following "
            f"the directions at the top of the system testfile "
        )

    training_config = {
        "TrainingJobName": training_job_name,
        "RoleArn": role_arn,
        "AlgorithmSpecification": {
            "TrainingImage": knn_image_uri,
            "TrainingInputMode": "File",
        },
        "HyperParameters": {
            "predictor_type": "classifier",
            "feature_dim": "4",
            "k": "3",
            "sample_size": str(TRAIN_DATA.count("\n") - 1),
        },
        "InputDataConfig": [
            {
                "ChannelName": "train",
                "CompressionType": "None",
                "ContentType": "text/csv",
                "DataSource": {
                    "S3DataSource": {
                        "S3DataDistributionType": "FullyReplicated",
                        "S3DataType": "S3Prefix",
                        "S3Uri": f"s3://{bucket_name}/{input_data_s3_key}/train.csv",
                    }
                },
            }
        ],
        "OutputDataConfig": {"S3OutputPath": f"s3://{bucket_name}/{training_output_s3_key}/"},
        "ResourceConfig": {
            "InstanceCount": 1,
            "InstanceType": "ml.m5.large",
            "VolumeSizeInGB": 1,
        },
        "StoppingCondition": {"MaxRuntimeInSeconds": 6 * 60},
    }

    model_config = {
        "ModelName": model_name,
        "ExecutionRoleArn": role_arn,
        "PrimaryContainer": {
            "Mode": "SingleModel",
            "Image": knn_image_uri,
            "ModelDataUrl": f"s3://{bucket_name}/{training_output_s3_key}/{training_job_name}/output/model.tar.gz",
        },
    }

    endpoint_config_config = {
        "EndpointConfigName": endpoint_config_job_name,
        "ProductionVariants": [
            {
                "VariantName": f"{env_id}-demo",
                "ModelName": model_name,
                "InstanceType": "ml.m5.large",
                "InitialInstanceCount": 1,
            },
        ],
    }

    deploy_endpoint_config = {
        "EndpointName": endpoint_name,
        "EndpointConfigName": endpoint_config_job_name,
    }

    ti.xcom_push(key="bucket_name", value=bucket_name)
    ti.xcom_push(key="input_data_s3_key", value=input_data_s3_key)
    ti.xcom_push(key="model_name", value=model_name)
    ti.xcom_push(key="endpoint_name", value=endpoint_name)
    ti.xcom_push(key="endpoint_config_job_name", value=endpoint_config_job_name)
    ti.xcom_push(key="training_config", value=training_config)
    ti.xcom_push(key="model_config", value=model_config)
    ti.xcom_push(key="endpoint_config_config", value=endpoint_config_config)
    ti.xcom_push(key="deploy_endpoint_config", value=deploy_endpoint_config)


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()

    test_setup = set_up(
        env_id=test_context[ENV_ID_KEY],
        role_arn=test_context[ROLE_ARN_KEY],
    )

    create_bucket = S3CreateBucketOperator(
        task_id="create_bucket",
        bucket_name=test_setup["bucket_name"],
    )

    upload_data = S3CreateObjectOperator(
        task_id="upload_data",
        s3_bucket=test_setup["bucket_name"],
        s3_key=f"{test_setup['input_data_s3_key']}/train.csv",
        data=TRAIN_DATA,
    )

    train_model = SageMakerTrainingOperator(
        task_id="train_model",
        config=test_setup["training_config"],
    )

    create_model = SageMakerModelOperator(
        task_id="create_model",
        config=test_setup["model_config"],
    )

    # [START howto_operator_sagemaker_endpoint_config]
    configure_endpoint = SageMakerEndpointConfigOperator(
        task_id="configure_endpoint",
        config=test_setup["endpoint_config_config"],
    )
    # [END howto_operator_sagemaker_endpoint_config]

    # [START howto_operator_sagemaker_endpoint]
    deploy_endpoint = SageMakerEndpointOperator(
        task_id="deploy_endpoint",
        config=test_setup["deploy_endpoint_config"],
    )
    # [END howto_operator_sagemaker_endpoint]

    # SageMakerEndpointOperator waits by default, setting as False to test the Sensor below.
    deploy_endpoint.wait_for_completion = False

    # [START howto_sensor_sagemaker_endpoint]
    await_endpoint = SageMakerEndpointSensor(
        task_id="await_endpoint",
        endpoint_name=test_setup["endpoint_name"],
    )
    # [END howto_sensor_sagemaker_endpoint]

    delete_model = SageMakerDeleteModelOperator(
        task_id="delete_model",
        trigger_rule=TriggerRule.ALL_DONE,
        config={"ModelName": test_setup["model_name"]},
    )

    delete_bucket = S3DeleteBucketOperator(
        task_id="delete_bucket",
        trigger_rule=TriggerRule.ALL_DONE,
        bucket_name=test_setup["bucket_name"],
        force_delete=True,
    )

    log_cleanup = prune_logs(
        [
            # Format: ('log group name', 'log stream prefix')
            ("/aws/sagemaker/TrainingJobs", test_context[ENV_ID_KEY]),
        ]
    )

    chain(
        # TEST SETUP
        test_context,
        test_setup,
        create_bucket,
        upload_data,
        # TEST BODY
        train_model,
        create_model,
        configure_endpoint,
        deploy_endpoint,
        await_endpoint,
        call_endpoint(test_setup["endpoint_name"]),
        # TEST TEARDOWN
        delete_endpoint_config(test_setup["endpoint_config_job_name"]),
        delete_endpoint(test_setup["endpoint_name"]),
        delete_model,
        delete_bucket,
        log_cleanup,
        archive_logs(f"/aws/sagemaker/Endpoints/{test_setup['endpoint_name']}"),
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import abc
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, TypeVar

from airflow.providers.amazon.aws.hooks.bedrock import BedrockAgentHook, BedrockHook
from airflow.providers.amazon.aws.sensors.base_aws import AwsBaseSensor
from airflow.providers.amazon.aws.triggers.bedrock import (
    BedrockBatchInferenceCompletedTrigger,
    BedrockBatchInferenceScheduledTrigger,
    BedrockCustomizeModelCompletedTrigger,
    BedrockIngestionJobTrigger,
    BedrockKnowledgeBaseActiveTrigger,
    BedrockProvisionModelThroughputCompletedTrigger,
)
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields
from airflow.providers.common.compat.sdk import AirflowException, conf

if TYPE_CHECKING:
    from airflow.providers.amazon.aws.triggers.bedrock import BedrockBaseBatchInferenceTrigger
    from airflow.sdk import Context


_GenericBedrockHook = TypeVar("_GenericBedrockHook", BedrockAgentHook, BedrockHook)


class BedrockBaseSensor(AwsBaseSensor[_GenericBedrockHook]):
    """
    General sensor behavior for Amazon Bedrock.

    Subclasses must implement following methods:
        - ``get_state()``

    Subclasses must set the following fields:
        - ``INTERMEDIATE_STATES``
        - ``FAILURE_STATES``
        - ``SUCCESS_STATES``
        - ``FAILURE_MESSAGE``

    :param deferrable: If True, the sensor will operate in deferrable mode. This mode requires aiobotocore
        module to be installed.
        (default: False, but can be overridden in config file by setting default_deferrable to True)
    """

    INTERMEDIATE_STATES: tuple[str, ...] = ()
    FAILURE_STATES: tuple[str, ...] = ()
    SUCCESS_STATES: tuple[str, ...] = ()
    FAILURE_MESSAGE = ""

    aws_hook_class: type[_GenericBedrockHook]
    ui_color = "#66c3ff"

    def __init__(
        self,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.deferrable = deferrable

    def poke(self, context: Context, **kwargs) -> bool:
        state = self.get_state()
        if state in self.FAILURE_STATES:
            raise AirflowException(self.FAILURE_MESSAGE)

        return state not in self.INTERMEDIATE_STATES

    @abc.abstractmethod
    def get_state(self) -> str:
        """Implement in subclasses."""


class BedrockCustomizeModelCompletedSensor(BedrockBaseSensor[BedrockHook]):
    """
    Poll the state of the model customization job until it reaches a terminal state; fails if the job fails.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:BedrockCustomizeModelCompletedSensor`

    :param job_name: The name of the Bedrock model customization job.

    :param deferrable: If True, the sensor will operate in deferrable mode. This mode requires aiobotocore
        module to be installed.
        (default: False, but can be overridden in config file by setting default_deferrable to True)
    :param poke_interval: Polling period in seconds to check for the status of the job. (default: 120)
    :param max_retries: Number of times before returning the current state. (default: 75)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    INTERMEDIATE_STATES: tuple[str, ...] = ("InProgress",)
    FAILURE_STATES: tuple[str, ...] = ("Failed", "Stopping", "Stopped")
    SUCCESS_STATES: tuple[str, ...] = ("Completed",)
    FAILURE_MESSAGE = "Bedrock model customization job sensor failed."

    aws_hook_class = BedrockHook

    template_fields: Sequence[str] = aws_template_fields("job_name")

    def __init__(
        self,
        *,
        job_name: str,
        max_retries: int = 75,
        poke_interval: int = 120,
        **kwargs: Any,
    ) -> None:
        super().__init__(**kwargs)
        self.poke_interval = poke_interval
        self.max_retries = max_retries
        self.job_name = job_name

    def execute(self, context: Context) -> Any:
        if self.deferrable:
            self.defer(
                trigger=BedrockCustomizeModelCompletedTrigger(
                    job_name=self.job_name,
                    waiter_delay=int(self.poke_interval),
                    waiter_max_attempts=self.max_retries,
                    aws_conn_id=self.aws_conn_id,
                ),
                method_name="poke",
            )
        else:
            super().execute(context=context)

    def get_state(self) -> str:
        return self.hook.conn.get_model_customization_job(jobIdentifier=self.job_name)["status"]


class BedrockProvisionModelThroughputCompletedSensor(BedrockBaseSensor[BedrockHook]):
    """
    Poll the provisioned model throughput job until it reaches a terminal state; fails if the job fails.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:BedrockProvisionModelThroughputCompletedSensor`

    :param model_id: The ARN or name of the provisioned throughput.

    :param deferrable: If True, the sensor will operate in deferrable more. This mode requires aiobotocore
        module to be installed.
        (default: False, but can be overridden in config file by setting default_deferrable to True)
    :param poke_interval: Polling period in seconds to check for the status of the job. (default: 60)
    :param max_retries: Number of times before returning the current state (default: 20)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    INTERMEDIATE_STATES: tuple[str, ...] = ("Creating", "Updating")
    FAILURE_STATES: tuple[str, ...] = ("Failed",)
    SUCCESS_STATES: tuple[str, ...] = ("InService",)
    FAILURE_MESSAGE = "Bedrock provision model throughput sensor failed."

    aws_hook_class = BedrockHook

    template_fields: Sequence[str] = aws_template_fields("model_id")

    def __init__(
        self,
        *,
        model_id: str,
        poke_interval: int = 60,
        max_retries: int = 20,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.poke_interval = poke_interval
        self.max_retries = max_retries
        self.model_id = model_id

    def get_state(self) -> str:
        return self.hook.conn.get_provisioned_model_throughput(provisionedModelId=self.model_id)["status"]

    def execute(self, context: Context) -> Any:
        if self.deferrable:
            self.defer(
                trigger=BedrockProvisionModelThroughputCompletedTrigger(
                    provisioned_model_id=self.model_id,
                    waiter_delay=int(self.poke_interval),
                    waiter_max_attempts=self.max_retries,
                    aws_conn_id=self.aws_conn_id,
                ),
                method_name="poke",
            )
        else:
            super().execute(context=context)


class BedrockKnowledgeBaseActiveSensor(BedrockBaseSensor[BedrockAgentHook]):
    """
    Poll the Knowledge Base status until it reaches a terminal state; fails if creation fails.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:BedrockKnowledgeBaseActiveSensor`

    :param knowledge_base_id: The unique identifier of the knowledge base for which to get information. (templated)

    :param deferrable: If True, the sensor will operate in deferrable more. This mode requires aiobotocore
        module to be installed.
        (default: False, but can be overridden in config file by setting default_deferrable to True)
    :param poke_interval: Polling period in seconds to check for the status of the job. (default: 5)
    :param max_retries: Number of times before returning the current state (default: 24)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    INTERMEDIATE_STATES: tuple[str, ...] = ("CREATING", "UPDATING")
    FAILURE_STATES: tuple[str, ...] = ("DELETING", "FAILED")
    SUCCESS_STATES: tuple[str, ...] = ("ACTIVE",)
    FAILURE_MESSAGE = "Bedrock Knowledge Base Active sensor failed."

    aws_hook_class = BedrockAgentHook

    template_fields: Sequence[str] = aws_template_fields("knowledge_base_id")

    def __init__(
        self,
        *,
        knowledge_base_id: str,
        poke_interval: int = 5,
        max_retries: int = 24,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.poke_interval = poke_interval
        self.max_retries = max_retries
        self.knowledge_base_id = knowledge_base_id

    def get_state(self) -> str:
        return self.hook.conn.get_knowledge_base(knowledgeBaseId=self.knowledge_base_id)["knowledgeBase"][
            "status"
        ]

    def execute(self, context: Context) -> Any:
        if self.deferrable:
            self.defer(
                trigger=BedrockKnowledgeBaseActiveTrigger(
                    knowledge_base_id=self.knowledge_base_id,
                    waiter_delay=int(self.poke_interval),
                    waiter_max_attempts=self.max_retries,
                    aws_conn_id=self.aws_conn_id,
                ),
                method_name="poke",
            )
        else:
            super().execute(context=context)


class BedrockIngestionJobSensor(BedrockBaseSensor[BedrockAgentHook]):
    """
    Poll the ingestion job status until it reaches a terminal state; fails if creation fails.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:BedrockIngestionJobSensor`

    :param knowledge_base_id: The unique identifier of the knowledge base for which to get information. (templated)
    :param data_source_id: The unique identifier of the data source in the ingestion job. (templated)
    :param ingestion_job_id: The unique identifier of the ingestion job. (templated)

    :param deferrable: If True, the sensor will operate in deferrable more. This mode requires aiobotocore
        module to be installed.
        (default: False, but can be overridden in config file by setting default_deferrable to True)
    :param poke_interval: Polling period in seconds to check for the status of the job. (default: 60)
    :param max_retries: Number of times before returning the current state (default: 10)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    INTERMEDIATE_STATES: tuple[str, ...] = ("STARTING", "IN_PROGRESS")
    FAILURE_STATES: tuple[str, ...] = ("FAILED",)
    SUCCESS_STATES: tuple[str, ...] = ("COMPLETE",)
    FAILURE_MESSAGE = "Bedrock ingestion job sensor failed."

    aws_hook_class = BedrockAgentHook

    template_fields: Sequence[str] = aws_template_fields(
        "knowledge_base_id", "data_source_id", "ingestion_job_id"
    )

    def __init__(
        self,
        *,
        knowledge_base_id: str,
        data_source_id: str,
        ingestion_job_id: str,
        poke_interval: int = 60,
        max_retries: int = 10,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.poke_interval = poke_interval
        self.max_retries = max_retries
        self.knowledge_base_id = knowledge_base_id
        self.data_source_id = data_source_id
        self.ingestion_job_id = ingestion_job_id

    def get_state(self) -> str:
        return self.hook.conn.get_ingestion_job(
            knowledgeBaseId=self.knowledge_base_id,
            ingestionJobId=self.ingestion_job_id,
            dataSourceId=self.data_source_id,
        )["ingestionJob"]["status"]

    def execute(self, context: Context) -> Any:
        if self.deferrable:
            self.defer(
                trigger=BedrockIngestionJobTrigger(
                    knowledge_base_id=self.knowledge_base_id,
                    ingestion_job_id=self.ingestion_job_id,
                    data_source_id=self.data_source_id,
                    waiter_delay=int(self.poke_interval),
                    waiter_max_attempts=self.max_retries,
                    aws_conn_id=self.aws_conn_id,
                ),
                method_name="poke",
            )
        else:
            super().execute(context=context)


class BedrockBatchInferenceSensor(BedrockBaseSensor[BedrockHook]):
    """
    Poll the batch inference job status until it reaches a terminal state; fails if creation fails.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:BedrockBatchInferenceSensor`

    :param job_arn: The Amazon Resource Name (ARN) of the batch inference job. (templated)
    :param success_state: A BedrockBatchInferenceSensor.TargetState; defaults to 'SCHEDULED' (templated)

    :param deferrable: If True, the sensor will operate in deferrable more. This mode requires aiobotocore
        module to be installed.
        (default: False, but can be overridden in config file by setting default_deferrable to True)
    :param poke_interval: Polling period in seconds to check for the status of the job. (default: 5)
    :param max_retries: Number of times before returning the current state (default: 24)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html
    """

    class SuccessState:
        """
        Target state for the BedrockBatchInferenceSensor.

        Bedrock adds batch inference jobs to a queue, and they may take some time to complete.
        If you want to wait for the job to complete, use TargetState.COMPLETED, but if you only want
        to wait until the service confirms that the job is in the queue, use TargetState.SCHEDULED.

        The normal successful progression of states is:
            Submitted > Validating > Scheduled > InProgress > PartiallyCompleted > Completed
        """

        SCHEDULED = "scheduled"
        COMPLETED = "completed"

    INTERMEDIATE_STATES: tuple[str, ...]  # Defined in __init__ based on target state
    FAILURE_STATES: tuple[str, ...] = ("Failed", "Stopped", "PartiallyCompleted", "Expired")
    SUCCESS_STATES: tuple[str, ...]  # Defined in __init__ based on target state
    FAILURE_MESSAGE = "Bedrock batch inference job sensor failed."
    INVALID_SUCCESS_STATE_MESSAGE = "success_state must be an instance of TargetState."

    aws_hook_class = BedrockHook

    template_fields: Sequence[str] = aws_template_fields("job_arn", "success_state")

    def __init__(
        self,
        *,
        job_arn: str,
        success_state: SuccessState | str = SuccessState.SCHEDULED,
        poke_interval: int = 120,
        max_retries: int = 75,
        **kwargs,
    ) -> None:
        super().__init__(**kwargs)
        self.poke_interval = poke_interval
        self.max_retries = max_retries
        self.job_arn = job_arn
        self.success_state = success_state

        base_success_states: tuple[str, ...] = ("Completed",)
        base_intermediate_states: tuple[str, ...] = ("Submitted", "InProgress", "Stopping", "Validating")
        scheduled_state = ("Scheduled",)
        self.trigger_class: type[BedrockBaseBatchInferenceTrigger]

        if self.success_state == BedrockBatchInferenceSensor.SuccessState.COMPLETED:
            intermediate_states = base_intermediate_states + scheduled_state
            success_states = base_success_states
            self.trigger_class = BedrockBatchInferenceCompletedTrigger
        elif self.success_state == BedrockBatchInferenceSensor.SuccessState.SCHEDULED:
            intermediate_states = base_intermediate_states
            success_states = base_success_states + scheduled_state
            self.trigger_class = BedrockBatchInferenceScheduledTrigger
        else:
            raise ValueError(
                "Success states for BedrockBatchInferenceSensor must be set using a BedrockBatchInferenceSensor.SuccessState"
            )

        BedrockBatchInferenceSensor.INTERMEDIATE_STATES = intermediate_states or base_intermediate_states
        BedrockBatchInferenceSensor.SUCCESS_STATES = success_states or base_success_states

    def get_state(self) -> str:
        return self.hook.conn.get_model_invocation_job(jobIdentifier=self.job_arn)["status"]

    def execute(self, context: Context) -> Any:
        if self.deferrable:
            self.defer(
                trigger=self.trigger_class(
                    job_arn=self.job_arn,
                    waiter_delay=int(self.poke_interval),
                    waiter_max_attempts=self.max_retries,
                    aws_conn_id=self.aws_conn_id,
                ),
                method_name="poke",
            )
        else:
            super().execute(context=context)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Tracking the state of Amazon EKS Clusters, Amazon EKS managed node groups, and AWS Fargate profiles."""

from __future__ import annotations

import warnings
from abc import abstractmethod
from collections.abc import Sequence
from typing import TYPE_CHECKING

from airflow.exceptions import AirflowProviderDeprecationWarning
from airflow.providers.amazon.aws.hooks.eks import (
    ClusterStates,
    EksHook,
    FargateProfileStates,
    NodegroupStates,
)
from airflow.providers.amazon.aws.sensors.base_aws import AwsBaseSensor
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields
from airflow.providers.common.compat.sdk import AirflowException

if TYPE_CHECKING:
    from airflow.sdk import Context


DEFAULT_CONN_ID = "aws_default"

CLUSTER_TERMINAL_STATES = frozenset({ClusterStates.ACTIVE, ClusterStates.FAILED, ClusterStates.NONEXISTENT})
FARGATE_TERMINAL_STATES = frozenset(
    {
        FargateProfileStates.ACTIVE,
        FargateProfileStates.CREATE_FAILED,
        FargateProfileStates.DELETE_FAILED,
        FargateProfileStates.NONEXISTENT,
    }
)
NODEGROUP_TERMINAL_STATES = frozenset(
    {
        NodegroupStates.ACTIVE,
        NodegroupStates.CREATE_FAILED,
        NodegroupStates.DELETE_FAILED,
        NodegroupStates.NONEXISTENT,
    }
)


class EksBaseSensor(AwsBaseSensor):
    """
    Base class to check various EKS states.

    Subclasses need to implement get_state and get_terminal_states methods.

    :param cluster_name: The name of the Cluster
    :param target_state: Will return successfully when that state is reached.
    :param target_state_type: The enum containing the states,
        will be used to convert the target state if it has to be converted from a string
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    """

    aws_hook_class = EksHook

    def __init__(
        self,
        *,
        cluster_name: str,
        target_state: ClusterStates | NodegroupStates | FargateProfileStates,
        target_state_type: type,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.cluster_name = cluster_name
        self.target_state = (
            target_state
            if isinstance(target_state, target_state_type)
            else target_state_type(str(target_state).upper())
        )

    def poke(self, context: Context) -> bool:
        state = self.get_state()
        self.log.info("Current state: %s", state)
        if state in (self.get_terminal_states() - {self.target_state}):
            # If we reach a terminal state which is not the target state
            raise AirflowException(
                f"Terminal state reached. Current state: {state}, Expected state: {self.target_state}"
            )
        return state == self.target_state

    @abstractmethod
    def get_state(self) -> ClusterStates | NodegroupStates | FargateProfileStates: ...

    @abstractmethod
    def get_terminal_states(self) -> frozenset: ...


class EksClusterStateSensor(EksBaseSensor):
    """
    Check the state of an Amazon EKS Cluster until it reaches the target state or another terminal state.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:EksClusterStateSensor`

    :param cluster_name: The name of the Cluster to watch. (templated)
    :param target_state: Target state of the Cluster. (templated)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    """

    template_fields: Sequence[str] = aws_template_fields("cluster_name", "target_state")
    ui_color = "#ff9900"
    ui_fgcolor = "#232F3E"

    def __init__(
        self,
        *,
        target_state: ClusterStates = ClusterStates.ACTIVE,
        region: str | None = None,
        **kwargs,
    ):
        if region is not None:
            warnings.warn(
                message="Parameter `region` is deprecated. Use the parameter `region_name` instead",
                category=AirflowProviderDeprecationWarning,
                stacklevel=2,
            )
            kwargs["region_name"] = region
        super().__init__(target_state=target_state, target_state_type=ClusterStates, **kwargs)

    def get_state(self) -> ClusterStates:
        return self.hook.get_cluster_state(clusterName=self.cluster_name)

    def get_terminal_states(self) -> frozenset:
        return CLUSTER_TERMINAL_STATES


class EksFargateProfileStateSensor(EksBaseSensor):
    """
    Check the state of an AWS Fargate profile until it reaches the target state or another terminal state.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:EksFargateProfileStateSensor`

    :param cluster_name: The name of the Cluster which the AWS Fargate profile is attached to. (templated)
    :param fargate_profile_name: The name of the Fargate profile to watch. (templated)
    :param target_state: Target state of the Fargate profile. (templated)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    """

    template_fields: Sequence[str] = aws_template_fields(
        "cluster_name", "fargate_profile_name", "target_state"
    )
    ui_color = "#ff9900"
    ui_fgcolor = "#232F3E"

    def __init__(
        self,
        *,
        fargate_profile_name: str,
        region: str | None = None,
        target_state: FargateProfileStates = FargateProfileStates.ACTIVE,
        **kwargs,
    ):
        if region is not None:
            warnings.warn(
                message="Parameter `region` is deprecated. Use the parameter `region_name` instead",
                category=AirflowProviderDeprecationWarning,
                stacklevel=2,
            )
            kwargs["region_name"] = region
        super().__init__(target_state=target_state, target_state_type=FargateProfileStates, **kwargs)
        self.fargate_profile_name = fargate_profile_name

    def get_state(self) -> FargateProfileStates:
        return self.hook.get_fargate_profile_state(
            clusterName=self.cluster_name, fargateProfileName=self.fargate_profile_name
        )

    def get_terminal_states(self) -> frozenset:
        return FARGATE_TERMINAL_STATES


class EksNodegroupStateSensor(EksBaseSensor):
    """
    Check the state of an EKS managed node group until it reaches the target state or another terminal state.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:EksNodegroupStateSensor`

    :param cluster_name: The name of the Cluster which the Nodegroup is attached to. (templated)
    :param nodegroup_name: The name of the Nodegroup to watch. (templated)
    :param target_state: Target state of the Nodegroup. (templated)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether or not to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    """

    template_fields: Sequence[str] = aws_template_fields("cluster_name", "nodegroup_name", "target_state")
    ui_color = "#ff9900"
    ui_fgcolor = "#232F3E"

    def __init__(
        self,
        *,
        nodegroup_name: str,
        target_state: NodegroupStates = NodegroupStates.ACTIVE,
        region: str | None = None,
        **kwargs,
    ):
        if region is not None:
            warnings.warn(
                message="Parameter `region` is deprecated. Use the parameter `region_name` instead",
                category=AirflowProviderDeprecationWarning,
                stacklevel=2,
            )
            kwargs["region_name"] = region
        super().__init__(target_state=target_state, target_state_type=NodegroupStates, **kwargs)
        self.nodegroup_name = nodegroup_name

    def get_state(self) -> NodegroupStates:
        return self.hook.get_nodegroup_state(clusterName=self.cluster_name, nodegroupName=self.nodegroup_name)

    def get_terminal_states(self) -> frozenset:
        return NODEGROUP_TERMINAL_STATES



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Generator
from unittest import mock

import pytest
from botocore.exceptions import WaiterError

from airflow.providers.amazon.aws.hooks.ssm import SsmHook
from airflow.providers.amazon.aws.operators.ssm import SsmGetCommandInvocationOperator, SsmRunCommandOperator

from unit.amazon.aws.utils.test_template_fields import validate_template_fields

COMMAND_ID = "test_command_id"
DOCUMENT_NAME = "test_ssm_custom_document"
INSTANCE_IDS = ["test_instance_id_1", "test_instance_id_2"]


class TestSsmRunCommandOperator:
    @pytest.fixture
    def mock_conn(self) -> Generator[SsmHook, None, None]:
        with mock.patch.object(SsmHook, "conn") as _conn:
            _conn.send_command.return_value = {
                "Command": {
                    "CommandId": COMMAND_ID,
                    "InstanceIds": INSTANCE_IDS,
                }
            }
            yield _conn

    def setup_method(self):
        self.operator = SsmRunCommandOperator(
            task_id="test_run_command_operator",
            document_name=DOCUMENT_NAME,
            run_command_kwargs={"InstanceIds": INSTANCE_IDS},
        )
        self.operator.defer = mock.MagicMock()

    @pytest.mark.parametrize(
        ("wait_for_completion", "deferrable"),
        [
            pytest.param(False, False, id="no_wait"),
            pytest.param(True, False, id="wait"),
            pytest.param(False, True, id="defer"),
        ],
    )
    @mock.patch.object(SsmHook, "get_waiter")
    def test_run_command_wait_combinations(self, mock_get_waiter, wait_for_completion, deferrable, mock_conn):
        self.operator.wait_for_completion = wait_for_completion
        self.operator.deferrable = deferrable

        command_id = self.operator.execute({})

        assert command_id == COMMAND_ID
        mock_conn.send_command.assert_called_once_with(DocumentName=DOCUMENT_NAME, InstanceIds=INSTANCE_IDS)
        assert mock_get_waiter.call_count == wait_for_completion
        assert self.operator.defer.call_count == deferrable

    def test_template_fields(self):
        validate_template_fields(self.operator)

    def test_deferrable_with_region(self, mock_conn):
        """Test that deferrable mode properly passes region and other AWS parameters to trigger."""
        self.operator.deferrable = True
        self.operator.region_name = "us-west-2"
        self.operator.verify = False
        self.operator.botocore_config = {"retries": {"max_attempts": 5}}

        command_id = self.operator.execute({})

        assert command_id == COMMAND_ID
        mock_conn.send_command.assert_called_once_with(DocumentName=DOCUMENT_NAME, InstanceIds=INSTANCE_IDS)

        # Verify defer was called with correct trigger parameters
        self.operator.defer.assert_called_once()
        call_args = self.operator.defer.call_args
        trigger = call_args[1]["trigger"]  # Get the trigger from kwargs

        # Verify the trigger has the correct parameters
        assert trigger.command_id == COMMAND_ID
        assert trigger.region_name == "us-west-2"
        assert trigger.verify is False
        assert trigger.botocore_config == {"retries": {"max_attempts": 5}}
        assert trigger.aws_conn_id == self.operator.aws_conn_id

    def test_operator_default_fails_on_nonzero_exit(self, mock_conn):
        """
        Test traditional mode where fail_on_nonzero_exit=True (default).

        Verifies that when fail_on_nonzero_exit is True (the default), the operator
        raises an exception when the waiter encounters a command failure.
        """
        self.operator.wait_for_completion = True

        # Mock waiter to raise WaiterError (simulating command failure)
        mock_waiter = mock.MagicMock()
        mock_waiter.wait.side_effect = WaiterError(
            name="command_executed",
            reason="Waiter encountered a terminal failure state",
            last_response={"Status": "Failed"},
        )

        with mock.patch.object(SsmHook, "get_waiter", return_value=mock_waiter):
            # Should raise WaiterError in traditional mode
            with pytest.raises(WaiterError):
                self.operator.execute({})

    def test_operator_enhanced_mode_tolerates_failed_status(self, mock_conn):
        """
        Test enhanced mode where fail_on_nonzero_exit=False tolerates Failed status.

        Verifies that when fail_on_nonzero_exit is False, the operator completes
        successfully even when the command returns a Failed status with non-zero exit code.
        """
        self.operator.wait_for_completion = True
        self.operator.fail_on_nonzero_exit = False

        # Mock waiter to raise WaiterError
        mock_waiter = mock.MagicMock()
        mock_waiter.wait.side_effect = WaiterError(
            name="command_executed",
            reason="Waiter encountered a terminal failure state",
            last_response={"Status": "Failed"},
        )

        # Mock get_command_invocation to return Failed status with exit code
        with (
            mock.patch.object(SsmHook, "get_waiter", return_value=mock_waiter),
            mock.patch.object(
                SsmHook, "get_command_invocation", return_value={"Status": "Failed", "ResponseCode": 1}
            ),
        ):
            # Should NOT raise in enhanced mode for Failed status
            command_id = self.operator.execute({})
            assert command_id == COMMAND_ID

    def test_operator_enhanced_mode_fails_on_timeout(self, mock_conn):
        """
        Test enhanced mode still fails on TimedOut status.

        Verifies that even when fail_on_nonzero_exit is False, the operator
        still raises an exception for AWS-level failures like TimedOut.
        """
        self.operator.wait_for_completion = True
        self.operator.fail_on_nonzero_exit = False

        # Mock waiter to raise WaiterError
        mock_waiter = mock.MagicMock()
        mock_waiter.wait.side_effect = WaiterError(
            name="command_executed",
            reason="Waiter encountered a terminal failure state",
            last_response={"Status": "TimedOut"},
        )

        # Mock get_command_invocation to return TimedOut status
        with (
            mock.patch.object(SsmHook, "get_waiter", return_value=mock_waiter),
            mock.patch.object(
                SsmHook, "get_command_invocation", return_value={"Status": "TimedOut", "ResponseCode": -1}
            ),
        ):
            # Should raise even in enhanced mode for TimedOut
            with pytest.raises(WaiterError):
                self.operator.execute({})

    def test_operator_enhanced_mode_fails_on_cancelled(self, mock_conn):
        """
        Test enhanced mode still fails on Cancelled status.

        Verifies that even when fail_on_nonzero_exit is False, the operator
        still raises an exception for AWS-level failures like Cancelled.
        """
        self.operator.wait_for_completion = True
        self.operator.fail_on_nonzero_exit = False

        # Mock waiter to raise WaiterError
        mock_waiter = mock.MagicMock()
        mock_waiter.wait.side_effect = WaiterError(
            name="command_executed",
            reason="Waiter encountered a terminal failure state",
            last_response={"Status": "Cancelled"},
        )

        # Mock get_command_invocation to return Cancelled status
        with (
            mock.patch.object(SsmHook, "get_waiter", return_value=mock_waiter),
            mock.patch.object(
                SsmHook, "get_command_invocation", return_value={"Status": "Cancelled", "ResponseCode": -1}
            ),
        ):
            # Should raise even in enhanced mode for Cancelled
            with pytest.raises(WaiterError):
                self.operator.execute({})

    @mock.patch("airflow.providers.amazon.aws.operators.ssm.SsmRunCommandTrigger")
    def test_operator_passes_parameter_to_trigger(self, mock_trigger_class, mock_conn):
        """
        Test that fail_on_nonzero_exit parameter is passed to trigger in deferrable mode.

        Verifies that when using deferrable mode, the fail_on_nonzero_exit parameter
        is correctly passed to the SsmRunCommandTrigger.
        """
        self.operator.deferrable = True
        self.operator.fail_on_nonzero_exit = False

        with mock.patch.object(self.operator, "defer") as mock_defer:
            command_id = self.operator.execute({})

            assert command_id == COMMAND_ID
            mock_conn.send_command.assert_called_once_with(
                DocumentName=DOCUMENT_NAME, InstanceIds=INSTANCE_IDS
            )

            # Verify defer was called
            mock_defer.assert_called_once()

            # Verify the trigger was instantiated with correct parameters
            mock_trigger_class.assert_called_once()
            call_kwargs = mock_trigger_class.call_args[1]

            assert call_kwargs["command_id"] == COMMAND_ID
            assert call_kwargs["fail_on_nonzero_exit"] is False


class TestSsmGetCommandInvocationOperator:
    @pytest.fixture
    def mock_hook(self) -> Generator[mock.MagicMock, None, None]:
        with mock.patch.object(SsmGetCommandInvocationOperator, "hook") as _hook:
            yield _hook

    def setup_method(self):
        self.command_id = "test-command-id-123"
        self.instance_id = "i-1234567890abcdef0"
        self.operator = SsmGetCommandInvocationOperator(
            task_id="test_get_command_invocation",
            command_id=self.command_id,
            instance_id=self.instance_id,
        )

    def test_execute_with_specific_instance(self, mock_hook):
        # Mock response for specific instance
        mock_invocation_details = {
            "Status": "Success",
            "ResponseCode": 0,
            "StandardOutputContent": "Hello World",
            "StandardErrorContent": "",
            "ExecutionStartDateTime": "2023-01-01T12:00:00Z",
            "ExecutionEndDateTime": "2023-01-01T12:00:05Z",
            "DocumentName": "AWS-RunShellScript",
            "Comment": "Test command",
        }
        mock_hook.get_command_invocation.return_value = mock_invocation_details

        result = self.operator.execute({})

        # Verify hook was called correctly
        mock_hook.get_command_invocation.assert_called_once_with(self.command_id, self.instance_id)

        # Verify returned data structure - should use standardized format with invocations array
        expected_result = {
            "command_id": self.command_id,
            "invocations": [
                {
                    "instance_id": self.instance_id,
                    "status": "Success",
                    "response_code": 0,
                    "standard_output": "Hello World",
                    "standard_error": "",
                    "execution_start_time": "2023-01-01T12:00:00Z",
                    "execution_end_time": "2023-01-01T12:00:05Z",
                    "document_name": "AWS-RunShellScript",
                    "comment": "Test command",
                }
            ],
        }
        assert result == expected_result

    def test_execute_all_instances(self, mock_hook):
        # Setup operator without instance_id to get all instances
        operator = SsmGetCommandInvocationOperator(
            task_id="test_get_all_invocations",
            command_id=self.command_id,
        )

        # Mock list_command_invocations response
        mock_invocations = [
            {"InstanceId": "i-111"},
            {"InstanceId": "i-222"},
        ]
        mock_hook.list_command_invocations.return_value = {"CommandInvocations": mock_invocations}

        # Mock get_command_invocation responses
        mock_invocation_details_1 = {
            "Status": "Success",
            "ResponseCode": 0,
            "StandardOutputContent": "Output 1",
            "StandardErrorContent": "",
            "ExecutionStartDateTime": "2023-01-01T12:00:00Z",
            "ExecutionEndDateTime": "2023-01-01T12:00:05Z",
            "DocumentName": "AWS-RunShellScript",
            "Comment": "",
        }
        mock_invocation_details_2 = {
            "Status": "Failed",
            "ResponseCode": 1,
            "StandardOutputContent": "",
            "StandardErrorContent": "Error occurred",
            "ExecutionStartDateTime": "2023-01-01T12:00:00Z",
            "ExecutionEndDateTime": "2023-01-01T12:00:10Z",
            "DocumentName": "AWS-RunShellScript",
            "Comment": "",
        }

        mock_hook.get_command_invocation.side_effect = [
            mock_invocation_details_1,
            mock_invocation_details_2,
        ]

        result = operator.execute({})

        # Verify hook calls
        mock_hook.list_command_invocations.assert_called_once_with(self.command_id)
        assert mock_hook.get_command_invocation.call_count == 2
        mock_hook.get_command_invocation.assert_any_call(self.command_id, "i-111")
        mock_hook.get_command_invocation.assert_any_call(self.command_id, "i-222")

        # Verify returned data structure
        expected_result = {
            "command_id": self.command_id,
            "invocations": [
                {
                    "instance_id": "i-111",
                    "status": "Success",
                    "response_code": 0,
                    "standard_output": "Output 1",
                    "standard_error": "",
                    "execution_start_time": "2023-01-01T12:00:00Z",
                    "execution_end_time": "2023-01-01T12:00:05Z",
                    "document_name": "AWS-RunShellScript",
                    "comment": "",
                },
                {
                    "instance_id": "i-222",
                    "status": "Failed",
                    "response_code": 1,
                    "standard_output": "",
                    "standard_error": "Error occurred",
                    "execution_start_time": "2023-01-01T12:00:00Z",
                    "execution_end_time": "2023-01-01T12:00:10Z",
                    "document_name": "AWS-RunShellScript",
                    "comment": "",
                },
            ],
        }
        assert result == expected_result

    def test_execute_all_instances_with_error(self, mock_hook):
        # Setup operator without instance_id
        operator = SsmGetCommandInvocationOperator(
            task_id="test_get_all_with_error",
            command_id=self.command_id,
        )

        # Mock list_command_invocations response
        mock_invocations = [{"InstanceId": "i-111"}]
        mock_hook.list_command_invocations.return_value = {"CommandInvocations": mock_invocations}

        # Mock get_command_invocation to raise an exception
        mock_hook.get_command_invocation.side_effect = Exception("API Error")

        result = operator.execute({})

        # Verify error handling
        expected_result = {
            "command_id": self.command_id,
            "invocations": [{"instance_id": "i-111", "error": "API Error"}],
        }
        assert result == expected_result

    def test_template_fields(self):
        validate_template_fields(self.operator)

    def test_exit_code_routing_use_case(self, mock_hook):
        """
        Test that demonstrates the exit code routing use case.

        This test verifies that SsmGetCommandInvocationOperator correctly retrieves
        exit codes and status information that can be used for workflow routing,
        particularly when used with SsmRunCommandOperator in enhanced mode
        (fail_on_nonzero_exit=False).
        """
        # Mock response with various exit codes that might be used for routing
        mock_invocation_details = {
            "Status": "Failed",  # Command failed but we want to route based on exit code
            "ResponseCode": 42,  # Custom exit code for specific routing logic
            "StandardOutputContent": "Partial success - some items processed",
            "StandardErrorContent": "Warning: 3 items skipped",
            "ExecutionStartDateTime": "2023-01-01T12:00:00Z",
            "ExecutionEndDateTime": "2023-01-01T12:00:05Z",
            "DocumentName": "AWS-RunShellScript",
            "Comment": "Data processing script",
        }
        mock_hook.get_command_invocation.return_value = mock_invocation_details

        result = self.operator.execute({})

        # Verify that response_code is available for routing decisions
        assert result["invocations"][0]["response_code"] == 42
        assert result["invocations"][0]["status"] == "Failed"

        # Verify that output is available for additional context
        assert "Partial success" in result["invocations"][0]["standard_output"]
        assert "Warning" in result["invocations"][0]["standard_error"]

        # This demonstrates that the operator provides all necessary information
        # for downstream tasks to make routing decisions based on exit codes,
        # which is the key use case for the enhanced mode feature.

    def test_multiple_exit_codes_for_routing(self, mock_hook):
        """
        Test retrieving multiple instances with different exit codes for routing.

        This demonstrates a common pattern where a command runs on multiple instances
        and downstream tasks need to route based on the exit codes from each instance.
        """
        operator = SsmGetCommandInvocationOperator(
            task_id="test_multi_instance_routing",
            command_id=self.command_id,
        )

        # Mock list_command_invocations response
        mock_invocations = [
            {"InstanceId": "i-success"},
            {"InstanceId": "i-partial"},
            {"InstanceId": "i-failed"},
        ]
        mock_hook.list_command_invocations.return_value = {"CommandInvocations": mock_invocations}

        # Mock different exit codes for routing scenarios
        mock_hook.get_command_invocation.side_effect = [
            {
                "Status": "Success",
                "ResponseCode": 0,  # Complete success
                "StandardOutputContent": "All items processed",
                "StandardErrorContent": "",
                "ExecutionStartDateTime": "2023-01-01T12:00:00Z",
                "ExecutionEndDateTime": "2023-01-01T12:00:05Z",
                "DocumentName": "AWS-RunShellScript",
                "Comment": "",
            },
            {
                "Status": "Failed",
                "ResponseCode": 2,  # Partial success - custom exit code
                "StandardOutputContent": "Some items processed",
                "StandardErrorContent": "Warning: partial completion",
                "ExecutionStartDateTime": "2023-01-01T12:00:00Z",
                "ExecutionEndDateTime": "2023-01-01T12:00:10Z",
                "DocumentName": "AWS-RunShellScript",
                "Comment": "",
            },
            {
                "Status": "Failed",
                "ResponseCode": 1,  # Complete failure
                "StandardOutputContent": "",
                "StandardErrorContent": "Error: operation failed",
                "ExecutionStartDateTime": "2023-01-01T12:00:00Z",
                "ExecutionEndDateTime": "2023-01-01T12:00:08Z",
                "DocumentName": "AWS-RunShellScript",
                "Comment": "",
            },
        ]

        result = operator.execute({})

        # Verify all exit codes are captured for routing logic
        assert len(result["invocations"]) == 3
        assert result["invocations"][0]["response_code"] == 0
        assert result["invocations"][1]["response_code"] == 2
        assert result["invocations"][2]["response_code"] == 1

        # This demonstrates that the operator can retrieve exit codes from multiple
        # instances, enabling complex routing logic based on the results from each instance.



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import boto3
import botocore
import pytest

from airflow.providers.amazon.aws.hooks.comprehend import ComprehendHook
from airflow.providers.amazon.aws.sensors.comprehend import (
    ComprehendCreateDocumentClassifierCompletedSensor,
    ComprehendStartPiiEntitiesDetectionJobCompletedSensor,
)


class TestComprehendCustomWaiters:
    def test_service_waiters(self):
        assert "pii_entities_detection_job_complete" in ComprehendHook().list_waiters()
        assert "create_document_classifier_complete" in ComprehendHook().list_waiters()


class TestComprehendCustomWaitersBase:
    @pytest.fixture(autouse=True)
    def mock_conn(self, monkeypatch):
        self.client = boto3.client("comprehend")
        monkeypatch.setattr(ComprehendHook, "conn", self.client)


class TestComprehendStartPiiEntitiesDetectionJobCompleteWaiter(TestComprehendCustomWaitersBase):
    WAITER_NAME = "pii_entities_detection_job_complete"

    @pytest.fixture
    def mock_get_job(self):
        with mock.patch.object(self.client, "describe_pii_entities_detection_job") as mock_getter:
            yield mock_getter

    @pytest.mark.parametrize("state", ComprehendStartPiiEntitiesDetectionJobCompletedSensor.SUCCESS_STATES)
    def test_pii_entities_detection_job_complete(self, state, mock_get_job):
        mock_get_job.return_value = {"PiiEntitiesDetectionJobProperties": {"JobStatus": state}}

        ComprehendHook().get_waiter(self.WAITER_NAME).wait(JobId="job_id")

    @pytest.mark.parametrize("state", ComprehendStartPiiEntitiesDetectionJobCompletedSensor.FAILURE_STATES)
    def test_pii_entities_detection_job_failed(self, state, mock_get_job):
        mock_get_job.return_value = {"PiiEntitiesDetectionJobProperties": {"JobStatus": state}}

        with pytest.raises(botocore.exceptions.WaiterError):
            ComprehendHook().get_waiter(self.WAITER_NAME).wait(JobId="job_id")

    def test_pii_entities_detection_job_wait(self, mock_get_job):
        wait = {"PiiEntitiesDetectionJobProperties": {"JobStatus": "IN_PROGRESS"}}
        success = {"PiiEntitiesDetectionJobProperties": {"JobStatus": "COMPLETED"}}
        mock_get_job.side_effect = [wait, wait, success]

        ComprehendHook().get_waiter(self.WAITER_NAME).wait(
            JobId="job_id", WaiterConfig={"Delay": 0.01, "MaxAttempts": 3}
        )


class TestComprehendDocumentClassifierCompleteWaiter(TestComprehendCustomWaitersBase):
    WAITER_NAME = "create_document_classifier_complete"

    @pytest.fixture
    def mock_describe_document_classifier(self):
        with mock.patch.object(self.client, "describe_document_classifier") as mock_getter:
            yield mock_getter

    @pytest.mark.parametrize("state", ComprehendCreateDocumentClassifierCompletedSensor.SUCCESS_STATES)
    def test_create_document_classifier_complete(self, state, mock_describe_document_classifier):
        mock_describe_document_classifier.return_value = {"DocumentClassifierProperties": {"Status": state}}

        ComprehendHook().get_waiter(self.WAITER_NAME).wait(DocumentClassifierArn="arn")

    @pytest.mark.parametrize("state", ComprehendCreateDocumentClassifierCompletedSensor.FAILURE_STATES)
    def test_create_document_classifier_failed(self, state, mock_describe_document_classifier):
        mock_describe_document_classifier.return_value = {"DocumentClassifierProperties": {"Status": state}}

        with pytest.raises(botocore.exceptions.WaiterError):
            ComprehendHook().get_waiter(self.WAITER_NAME).wait(DocumentClassifierArn="arn")

    def test_create_document_classifier_wait(self, mock_describe_document_classifier):
        wait = {"DocumentClassifierProperties": {"Status": "TRAINING"}}
        success = {"DocumentClassifierProperties": {"Status": "TRAINED"}}
        mock_describe_document_classifier.side_effect = [wait, wait, success]

        ComprehendHook().get_waiter(self.WAITER_NAME).wait(
            DocumentClassifierArn="arn", WaiterConfig={"Delay": 0.01, "MaxAttempts": 3}
        )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import re
from typing import TYPE_CHECKING

from airflow.providers.amazon.aws.triggers.sqs import SqsSensorTrigger
from airflow.providers.common.compat.sdk import AirflowOptionalProviderFeatureException

try:
    from airflow.providers.common.messaging.providers.base_provider import BaseMessageQueueProvider
except ImportError:
    raise AirflowOptionalProviderFeatureException(
        "This feature requires the 'common.messaging' provider to be installed in version >= 1.0.1."
    )

if TYPE_CHECKING:
    from airflow.triggers.base import BaseEventTrigger

# [START queue_regexp]
QUEUE_REGEXP = r"^https://sqs\.[^.]+\.amazonaws\.com/[0-9]+/.+"
# [END queue_regexp]


class SqsMessageQueueProvider(BaseMessageQueueProvider):
    """
    Configuration for SQS integration with common-messaging.

    [START sqs_message_queue_provider_description]

    * It uses ``sqs`` as scheme for identifying SQS queues.
    * For parameter definitions take a look at :class:`~airflow.providers.amazon.aws.triggers.sqs.SqsSensorTrigger`.

    .. code-block:: python

        from airflow.providers.common.messaging.triggers.msg_queue import MessageQueueTrigger
        from airflow.sdk import Asset, AssetWatcher

        trigger = MessageQueueTrigger(
            scheme="sqs",
            # Additional AWS SqsSensorTrigger parameters as needed
            sqs_queue="https://sqs.us-east-1.amazonaws.com/123456789012/my-queue",
            aws_conn_id="aws_default",
        )

        asset = Asset("sqs_queue_asset", watchers=[AssetWatcher(name="sqs_watcher", trigger=trigger)])

    For a complete example, see:
    :mod:`tests.system.amazon.aws.example_dag_sqs_message_queue_trigger`

    [END sqs_message_queue_provider_description]
    """

    scheme = "sqs"

    def queue_matches(self, queue: str) -> bool:
        return bool(re.match(QUEUE_REGEXP, queue))

    def trigger_class(self) -> type[BaseEventTrigger]:
        return SqsSensorTrigger

    def trigger_kwargs(self, queue: str, **kwargs) -> dict:
        return {
            "sqs_queue": queue,
        }



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock

import pytest

from airflow.providers.amazon.aws.exceptions import EcsOperatorError
from airflow.providers.amazon.aws.hooks.ecs import EcsHook, should_retry

DEFAULT_CONN_ID: str = "aws_default"
REGION: str = "us-east-1"


@pytest.fixture
def mock_conn():
    with mock.patch.object(EcsHook, "conn") as _conn:
        yield _conn


class TestEksHooks:
    def test_hook(self) -> None:
        hook = EcsHook(region_name=REGION)
        assert hook.conn is not None
        assert hook.aws_conn_id == DEFAULT_CONN_ID
        assert hook.region_name == REGION

    def test_get_cluster_state(self, mock_conn) -> None:
        mock_conn.describe_clusters.return_value = {"clusters": [{"status": "ACTIVE"}]}
        assert EcsHook().get_cluster_state(cluster_name="cluster_name") == "ACTIVE"

    def test_get_task_definition_state(self, mock_conn) -> None:
        mock_conn.describe_task_definition.return_value = {"taskDefinition": {"status": "ACTIVE"}}
        assert EcsHook().get_task_definition_state(task_definition="task_name") == "ACTIVE"

    def test_get_task_state(self, mock_conn) -> None:
        mock_conn.describe_tasks.return_value = {"tasks": [{"lastStatus": "ACTIVE"}]}
        assert EcsHook().get_task_state(cluster="cluster_name", task="task_name") == "ACTIVE"


class TestShouldRetry:
    def test_return_true_on_valid_reason(self):
        assert should_retry(EcsOperatorError([{"reason": "RESOURCE:MEMORY"}], "Foo"))

    def test_return_false_on_invalid_reason(self):
        assert not should_retry(EcsOperatorError([{"reason": "CLUSTER_NOT_FOUND"}], "Foo"))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import functools
import time
from collections.abc import Callable
from typing import ParamSpec, TypeVar

from airflow.providers.amazon.aws.hooks.base_aws import AwsBaseHook
from airflow.providers.common.compat.sdk import AirflowException

PS = ParamSpec("PS")
RT = TypeVar("RT")


def only_client_type(func: Callable[PS, RT]) -> Callable[PS, RT]:
    @functools.wraps(func)
    def checker(*args, **kwargs) -> RT:
        if args[0]._api_type == "client_type":
            return func(*args, **kwargs)

        ec2_doc_link = "https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html"
        raise AirflowException(
            f"""
            This method is only callable when using client_type API for interacting with EC2.
            Create the EC2Hook object as follows to use this method

            ec2 = EC2Hook(api_type="client_type")

            Read following for details on client_type and resource_type APIs:
            1. {ec2_doc_link}#client
            2. {ec2_doc_link}#service-resource
            """
        )

    return checker


class EC2Hook(AwsBaseHook):
    """
    Interact with Amazon Elastic Compute Cloud (EC2).

    Provide thick wrapper around :external+boto3:py:class:`boto3.client("ec2") <EC2.Client>`
    or :external+boto3:py:class:`boto3.resource("ec2") <EC2.ServiceResource>`.

    :param api_type: If set to ``client_type`` then hook use ``boto3.client("ec2")`` capabilities,
        If set to ``resource_type`` then hook use ``boto3.resource("ec2")`` capabilities.

    Additional arguments (such as ``aws_conn_id``) may be specified and
    are passed down to the underlying AwsBaseHook.

    .. seealso::
        - :class:`airflow.providers.amazon.aws.hooks.base_aws.AwsBaseHook`
    """

    API_TYPES = frozenset({"resource_type", "client_type"})

    def __init__(self, api_type="resource_type", *args, **kwargs) -> None:
        if api_type not in self.API_TYPES:
            raise AirflowException("api_type can only be one of %s", self.API_TYPES)

        kwargs[api_type] = "ec2"

        self._api_type = api_type

        super().__init__(*args, **kwargs)

    def get_instance(self, instance_id: str, filters: list | None = None):
        """
        Get EC2 instance by id and return it.

        :param instance_id: id of the AWS EC2 instance
        :param filters: List of filters to specify instances to get
        :return: Instance object
        """
        if self._api_type == "client_type":
            return self.get_instances(filters=filters, instance_ids=[instance_id])[0]

        return self.conn.Instance(id=instance_id)

    @only_client_type
    def stop_instances(self, instance_ids: list) -> dict:
        """
        Stop instances with given ids.

        :param instance_ids: List of instance ids to stop
        :return: Dict with key `StoppingInstances` and value as list of instances being stopped
        """
        self.log.info("Stopping instances: %s", instance_ids)

        return self.conn.stop_instances(InstanceIds=instance_ids)

    @only_client_type
    def start_instances(self, instance_ids: list) -> dict:
        """
        Start instances with given ids.

        :param instance_ids: List of instance ids to start
        :return: Dict with key `StartingInstances` and value as list of instances being started
        """
        self.log.info("Starting instances: %s", instance_ids)

        return self.conn.start_instances(InstanceIds=instance_ids)

    @only_client_type
    def terminate_instances(self, instance_ids: list) -> dict:
        """
        Terminate instances with given ids.

        :param instance_ids: List of instance ids to terminate
        :return: Dict with key `TerminatingInstances` and value as list of instances being terminated
        """
        self.log.info("Terminating instances: %s", instance_ids)

        return self.conn.terminate_instances(InstanceIds=instance_ids)

    @only_client_type
    def describe_instances(self, filters: list | None = None, instance_ids: list | None = None):
        """
        Describe EC2 instances, optionally applying filters and selective instance ids.

        :param filters: List of filters to specify instances to describe
        :param instance_ids: List of instance IDs to describe
        :return: Response from EC2 describe_instances API
        """
        filters = filters or []
        instance_ids = instance_ids or []

        self.log.info("Filters provided: %s", filters)
        self.log.info("Instance ids provided: %s", instance_ids)

        return self.conn.describe_instances(Filters=filters, InstanceIds=instance_ids)

    @only_client_type
    def get_instances(self, filters: list | None = None, instance_ids: list | None = None) -> list:
        """
        Get list of instance details, optionally applying filters and selective instance ids.

        :param instance_ids: List of ids to get instances for
        :param filters: List of filters to specify instances to get
        :return: List of instances
        """
        description = self.describe_instances(filters=filters, instance_ids=instance_ids)

        return [
            instance for reservation in description["Reservations"] for instance in reservation["Instances"]
        ]

    @only_client_type
    def get_instance_ids(self, filters: list | None = None) -> list:
        """
        Get list of instance ids, optionally applying filters to fetch selective instances.

        :param filters: List of filters to specify instances to get
        :return: List of instance ids
        """
        return [instance["InstanceId"] for instance in self.get_instances(filters=filters)]

    async def get_instance_state_async(self, instance_id: str) -> str:
        async with await self.get_async_conn() as client:
            response = await client.describe_instances(InstanceIds=[instance_id])
            return response["Reservations"][0]["Instances"][0]["State"]["Name"]

    def get_instance_state(self, instance_id: str) -> str:
        """
        Get EC2 instance state by id and return it.

        :param instance_id: id of the AWS EC2 instance
        :return: current state of the instance
        """
        if self._api_type == "client_type":
            return self.get_instances(instance_ids=[instance_id])[0]["State"]["Name"]

        return self.get_instance(instance_id=instance_id).state["Name"]

    def wait_for_state(self, instance_id: str, target_state: str, check_interval: float) -> None:
        """
        Wait EC2 instance until its state is equal to the target_state.

        :param instance_id: id of the AWS EC2 instance
        :param target_state: target state of instance
        :param check_interval: time in seconds that the job should wait in
            between each instance state checks until operation is completed
        :return: None
        """
        instance_state = self.get_instance_state(instance_id=instance_id)

        while instance_state != target_state:
            time.sleep(check_interval)
            instance_state = self.get_instance_state(instance_id=instance_id)

            self.log.info(
                "instance state: %s. Same as target: %s", instance_state, instance_state == target_state
            )



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

import boto3

from airflow.providers.amazon.aws.operators.datasync import DataSyncOperator
from airflow.providers.amazon.aws.operators.s3 import S3CreateBucketOperator, S3DeleteBucketOperator

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import ENV_ID_KEY, SystemTestContextBuilder

DAG_ID = "example_datasync"

# Externally fetched variables:
ROLE_ARN_KEY = "ROLE_ARN"

sys_test_context_task = SystemTestContextBuilder().add_variable(ROLE_ARN_KEY).build()


def get_s3_bucket_arn(bucket_name):
    return f"arn:aws:s3:::{bucket_name}"


def create_location(bucket_name, role_arn):
    client = boto3.client("datasync")
    response = client.create_location_s3(
        Subdirectory="test",
        S3BucketArn=get_s3_bucket_arn(bucket_name),
        S3Config={
            "BucketAccessRoleArn": role_arn,
        },
    )
    return response["LocationArn"]


@task
def create_source_location(bucket_source, role_arn):
    return create_location(bucket_source, role_arn)


@task
def create_destination_location(bucket_destination, role_arn):
    return create_location(bucket_destination, role_arn)


@task
def create_task(**kwargs):
    client = boto3.client("datasync")
    response = client.create_task(
        SourceLocationArn=kwargs["ti"].xcom_pull("create_source_location"),
        DestinationLocationArn=kwargs["ti"].xcom_pull("create_destination_location"),
    )
    return response["TaskArn"]


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_task(task_arn):
    client = boto3.client("datasync")
    client.delete_task(
        TaskArn=task_arn,
    )


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_task_created_by_operator(**kwargs):
    client = boto3.client("datasync")
    client.delete_task(
        TaskArn=kwargs["ti"].xcom_pull("create_and_execute_task")["TaskArn"],
    )


@task(trigger_rule=TriggerRule.ALL_DONE)
def list_locations(bucket_source, bucket_destination):
    client = boto3.client("datasync")
    return client.list_locations(
        Filters=[
            {
                "Name": "LocationUri",
                "Values": [
                    f"s3://{bucket_source}/test/",
                    f"s3://{bucket_destination}/test/",
                    f"s3://{bucket_source}/test_create/",
                    f"s3://{bucket_destination}/test_create/",
                ],
                "Operator": "In",
            }
        ]
    )


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_locations(locations):
    client = boto3.client("datasync")
    for location in locations["Locations"]:
        client.delete_location(
            LocationArn=location["LocationArn"],
        )


with DAG(
    DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()

    s3_bucket_source: str = f"{test_context[ENV_ID_KEY]}-datasync-bucket-source"
    s3_bucket_destination: str = f"{test_context[ENV_ID_KEY]}-datasync-bucket-destination"

    create_s3_bucket_source = S3CreateBucketOperator(
        task_id="create_s3_bucket_source", bucket_name=s3_bucket_source
    )

    create_s3_bucket_destination = S3CreateBucketOperator(
        task_id="create_s3_bucket_destination", bucket_name=s3_bucket_destination
    )

    source_location = create_source_location(s3_bucket_source, test_context[ROLE_ARN_KEY])
    destination_location = create_destination_location(s3_bucket_destination, test_context[ROLE_ARN_KEY])

    created_task_arn = create_task()

    # [START howto_operator_datasync_specific_task]
    # Execute a specific task
    execute_task_by_arn = DataSyncOperator(
        task_id="execute_task_by_arn",
        task_arn=created_task_arn,
    )
    # [END howto_operator_datasync_specific_task]

    # DataSyncOperator waits by default, setting as False to test the Sensor below.
    execute_task_by_arn.wait_for_completion = False

    # [START howto_operator_datasync_search_task]
    # Search and execute a task
    execute_task_by_locations = DataSyncOperator(
        task_id="execute_task_by_locations",
        source_location_uri=f"s3://{s3_bucket_source}/test",
        destination_location_uri=f"s3://{s3_bucket_destination}/test",
        # Only transfer files from /test/subdir folder
        task_execution_kwargs={
            "Includes": [{"FilterType": "SIMPLE_PATTERN", "Value": "/test/subdir"}],
        },
    )
    # [END howto_operator_datasync_search_task]

    # DataSyncOperator waits by default, setting as False to test the Sensor below.
    execute_task_by_locations.wait_for_completion = False

    # [START howto_operator_datasync_create_task]
    # Create a task (the task does not exist)
    create_and_execute_task = DataSyncOperator(
        task_id="create_and_execute_task",
        source_location_uri=f"s3://{s3_bucket_source}/test_create",
        destination_location_uri=f"s3://{s3_bucket_destination}/test_create",
        create_task_kwargs={"Name": "Created by Airflow"},
        create_source_location_kwargs={
            "Subdirectory": "test_create",
            "S3BucketArn": get_s3_bucket_arn(s3_bucket_source),
            "S3Config": {
                "BucketAccessRoleArn": test_context[ROLE_ARN_KEY],
            },
        },
        create_destination_location_kwargs={
            "Subdirectory": "test_create",
            "S3BucketArn": get_s3_bucket_arn(s3_bucket_destination),
            "S3Config": {
                "BucketAccessRoleArn": test_context[ROLE_ARN_KEY],
            },
        },
        delete_task_after_execution=False,
    )
    # [END howto_operator_datasync_create_task]

    # DataSyncOperator waits by default, setting as False to test the Sensor below.
    create_and_execute_task.wait_for_completion = False

    locations_task = list_locations(s3_bucket_source, s3_bucket_destination)
    delete_locations_task = delete_locations(locations_task)

    delete_s3_bucket_source = S3DeleteBucketOperator(
        task_id="delete_s3_bucket_source",
        bucket_name=s3_bucket_source,
        force_delete=True,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    delete_s3_bucket_destination = S3DeleteBucketOperator(
        task_id="delete_s3_bucket_destination",
        bucket_name=s3_bucket_destination,
        force_delete=True,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    chain(
        # TEST SETUP
        test_context,
        create_s3_bucket_source,
        create_s3_bucket_destination,
        source_location,
        destination_location,
        created_task_arn,
        # TEST BODY
        execute_task_by_arn,
        execute_task_by_locations,
        create_and_execute_task,
        # TEST TEARDOWN
        delete_task(created_task_arn),
        delete_task_created_by_operator(),
        locations_task,
        delete_locations_task,
        delete_s3_bucket_source,
        delete_s3_bucket_destination,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import json

from moto import mock_aws

from airflow.providers.amazon.aws.hooks.cloud_formation import CloudFormationHook


@mock_aws
class TestCloudFormationHook:
    def setup_method(self, _):
        self.hook = CloudFormationHook(aws_conn_id="aws_default")

    def create_stack(self, stack_name):
        timeout = 15
        template_body = json.dumps(
            {
                "Resources": {
                    "myResource": {
                        "Type": "AWS::EC2::VPC",
                        "Properties": {
                            "CidrBlock": {"Ref": "VPCCidr"},
                            "Tags": [{"Key": "Name", "Value": "Primary_CF_VPC"}],
                        },
                    }
                },
                "Parameters": {
                    "VPCCidr": {
                        "Type": "String",
                        "Default": "10.0.0.0/16",
                        "Description": "Enter the CIDR block for the VPC. Default is 10.0.0.0/16.",
                    }
                },
            }
        )

        self.hook.create_stack(
            stack_name=stack_name,
            cloudformation_parameters={
                "TimeoutInMinutes": timeout,
                "TemplateBody": template_body,
                "Parameters": [{"ParameterKey": "VPCCidr", "ParameterValue": "10.0.0.0/16"}],
            },
        )

    def test_get_conn_returns_a_boto3_connection(self):
        assert self.hook.get_conn().describe_stacks() is not None

    def test_get_stack_status(self):
        stack_name = "my_test_get_stack_status_stack"

        stack_status = self.hook.get_stack_status(stack_name=stack_name)
        assert stack_status is None

        self.create_stack(stack_name)
        stack_status = self.hook.get_stack_status(stack_name=stack_name)
        assert stack_status == "CREATE_COMPLETE", "Incorrect stack status returned."

    def test_create_stack(self):
        stack_name = "my_test_create_stack_stack"
        self.create_stack(stack_name)

        stacks = self.hook.get_conn().describe_stacks()["Stacks"]
        assert len(stacks) > 0, "CloudFormation should have stacks"

        matching_stacks = [x for x in stacks if x["StackName"] == stack_name]
        assert len(matching_stacks) == 1, f"stack with name {stack_name} should exist"

        stack = matching_stacks[0]
        assert stack["StackStatus"] == "CREATE_COMPLETE", "Stack should be in status CREATE_COMPLETE"

    def test_delete_stack(self):
        stack_name = "my_test_delete_stack_stack"
        self.create_stack(stack_name)

        self.hook.delete_stack(stack_name=stack_name)

        stacks = self.hook.get_conn().describe_stacks()["Stacks"]
        matching_stacks = [x for x in stacks if x["StackName"] == stack_name]
        assert not matching_stacks, f"stack with name {stack_name} should not exist"



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import logging
from typing import Any
from unittest import mock
from unittest.mock import AsyncMock

import pytest
from botocore.exceptions import WaiterError

from airflow.providers.amazon.aws.utils.waiter_with_logging import _LazyStatusFormatter, async_wait, wait
from airflow.providers.common.compat.sdk import AirflowException


def generate_response(state: str) -> dict[str, Any]:
    return {
        "Status": {
            "State": state,
        },
    }


class TestWaiter:
    @mock.patch("time.sleep")
    def test_wait(self, mock_sleep):
        mock_sleep.return_value = True
        mock_waiter = mock.MagicMock()
        error = WaiterError(
            name="test_waiter",
            reason="test_reason",
            last_response=generate_response("Pending"),
        )
        mock_waiter.wait.side_effect = [error, error, True]
        wait(
            waiter=mock_waiter,
            waiter_delay=123,
            waiter_max_attempts=456,
            args={"test_arg": "test_value"},
            failure_message="test failure message",
            status_message="test status message",
            status_args=["Status.State"],
        )

        mock_waiter.wait.assert_called_with(
            **{"test_arg": "test_value"},
            WaiterConfig={
                "MaxAttempts": 1,
            },
        )
        assert mock_waiter.wait.call_count == 3
        mock_sleep.assert_called_with(123)

    @pytest.mark.asyncio
    async def test_async_wait(self):
        mock_waiter = mock.MagicMock()
        error = WaiterError(
            name="test_waiter",
            reason="test_reason",
            last_response=generate_response("Pending"),
        )
        mock_waiter.wait = AsyncMock()
        mock_waiter.wait.side_effect = [error, error, True]

        await async_wait(
            waiter=mock_waiter,
            waiter_delay=0,
            waiter_max_attempts=456,
            args={"test_arg": "test_value"},
            failure_message="test failure message",
            status_message="test status message",
            status_args=["Status.State"],
        )

        mock_waiter.wait.assert_called_with(
            **{"test_arg": "test_value"},
            WaiterConfig={
                "MaxAttempts": 1,
            },
        )
        assert mock_waiter.wait.call_count == 3

    @pytest.mark.asyncio
    async def test_async_wait_with_unknown_failure(self):
        mock_waiter = mock.MagicMock()
        service_exception = WaiterError(
            name="test_waiter",
            reason="An error occurred",
            last_response={
                "Error": {
                    "Message": "Not authorized to perform: states:DescribeExecution on resource",
                    "Code": "AccessDeniedException",
                }
            },
        )
        mock_waiter.wait = AsyncMock()
        mock_waiter.wait.side_effect = [service_exception]
        with pytest.raises(AirflowException) as exc:
            await async_wait(
                waiter=mock_waiter,
                waiter_delay=0,
                waiter_max_attempts=456,
                args={"test_arg": "test_value"},
                failure_message="test failure message",
                status_message="test status message",
                status_args=["Status.State"],
            )

        mock_waiter.wait.assert_called_with(
            **{"test_arg": "test_value"},
            WaiterConfig={
                "MaxAttempts": 1,
            },
        )
        assert "An error occurred" in str(exc)
        assert mock_waiter.wait.call_count == 1

    @mock.patch("time.sleep")
    def test_wait_max_attempts_exceeded(self, mock_sleep):
        mock_sleep.return_value = True
        mock_waiter = mock.MagicMock()
        error = WaiterError(
            name="test_waiter",
            reason="test_reason",
            last_response=generate_response("Pending"),
        )
        mock_waiter.wait.side_effect = [error, error, error]
        with pytest.raises(AirflowException) as exc:
            wait(
                waiter=mock_waiter,
                waiter_delay=123,
                waiter_max_attempts=2,
                args={"test_arg": "test_value"},
                failure_message="test failure message",
                status_message="test status message",
                status_args=["Status.State"],
            )
        assert "Waiter error: max attempts reached" in str(exc)
        mock_waiter.wait.assert_called_with(
            **{"test_arg": "test_value"},
            WaiterConfig={
                "MaxAttempts": 1,
            },
        )

        assert mock_waiter.wait.call_count == 2
        mock_sleep.assert_called_with(123)

    @mock.patch("time.sleep")
    def test_wait_with_failure(self, mock_sleep):
        mock_sleep.return_value = True
        mock_waiter = mock.MagicMock()
        error = WaiterError(
            name="test_waiter",
            reason="test_reason",
            last_response=generate_response("Pending"),
        )
        failure_error = WaiterError(
            name="test_waiter",
            reason="terminal failure in waiter",
            last_response=generate_response("Failure"),
        )
        mock_waiter.wait.side_effect = [error, error, error, failure_error]

        with pytest.raises(AirflowException) as exc:
            wait(
                waiter=mock_waiter,
                waiter_delay=123,
                waiter_max_attempts=10,
                args={"test_arg": "test_value"},
                failure_message="test failure message",
                status_message="test status message",
                status_args=["Status.State"],
            )

        assert "test failure message" in str(exc)
        mock_waiter.wait.assert_called_with(
            **{"test_arg": "test_value"},
            WaiterConfig={
                "MaxAttempts": 1,
            },
        )
        assert mock_waiter.wait.call_count == 4

    @mock.patch("time.sleep")
    def test_wait_with_unknown_failure(self, mock_sleep):
        mock_sleep.return_value = True
        mock_waiter = mock.MagicMock()
        service_exception = WaiterError(
            name="test_waiter",
            reason="An error occurred",
            last_response={
                "Error": {
                    "Message": "Not authorized to perform: states:DescribeExecution on resource",
                    "Code": "AccessDeniedException",
                }
            },
        )
        mock_waiter.wait.side_effect = [service_exception]

        with pytest.raises(AirflowException) as exc:
            wait(
                waiter=mock_waiter,
                waiter_delay=123,
                waiter_max_attempts=10,
                args={"test_arg": "test_value"},
                failure_message="test failure message",
                status_message="test status message",
                status_args=["Status.State"],
            )

        assert "An error occurred" in str(exc)
        mock_waiter.wait.assert_called_with(
            **{"test_arg": "test_value"},
            WaiterConfig={
                "MaxAttempts": 1,
            },
        )
        assert mock_waiter.wait.call_count == 1

    @mock.patch("time.sleep")
    def test_wait_with_list_response(self, mock_sleep):
        mock_sleep.return_value = True
        mock_waiter = mock.MagicMock()
        error = WaiterError(
            name="test_waiter",
            reason="test_reason",
            last_response={
                "Clusters": [
                    {
                        "Status": "Pending",
                    },
                    {
                        "Status": "Pending",
                    },
                ]
            },
        )
        mock_waiter.wait.side_effect = [error, error, True]
        wait(
            waiter=mock_waiter,
            waiter_delay=123,
            waiter_max_attempts=456,
            args={"test_arg": "test_value"},
            failure_message="test failure message",
            status_message="test status message",
            status_args=["Clusters[0].Status"],
        )

        mock_waiter.wait.assert_called_with(
            **{"test_arg": "test_value"},
            WaiterConfig={
                "MaxAttempts": 1,
            },
        )
        mock_waiter.wait.call_count == 3
        mock_sleep.assert_called_with(123)

    @mock.patch("time.sleep")
    def test_wait_with_incorrect_args(self, mock_sleep):
        mock_sleep.return_value = True
        mock_waiter = mock.MagicMock()
        error = WaiterError(
            name="test_waiter",
            reason="test_reason",
            last_response={
                "Clusters": [
                    {
                        "Status": "Pending",
                    },
                    {
                        "Status": "Pending",
                    },
                ]
            },
        )
        mock_waiter.wait.side_effect = [error, error, True]
        wait(
            waiter=mock_waiter,
            waiter_delay=123,
            waiter_max_attempts=456,
            args={"test_arg": "test_value"},
            failure_message="test failure message",
            status_message="test status message",
            status_args=["Clusters[0].State"],  # this does not exist in the response
        )

        mock_waiter.wait.assert_called_with(
            **{"test_arg": "test_value"},
            WaiterConfig={
                "MaxAttempts": 1,
            },
        )
        assert mock_waiter.wait.call_count == 3
        mock_sleep.assert_called_with(123)

    @mock.patch("time.sleep")
    def test_wait_with_multiple_args(self, mock_sleep):
        mock_sleep.return_value = True
        mock_waiter = mock.MagicMock()
        error = WaiterError(
            name="test_waiter",
            reason="test_reason",
            last_response={
                "Clusters": [
                    {
                        "Status": "Pending",
                        "StatusDetails": "test_details",
                        "ClusterName": "test_name",
                    },
                ]
            },
        )
        mock_waiter.wait.side_effect = [error, error, True]
        wait(
            waiter=mock_waiter,
            waiter_delay=123,
            waiter_max_attempts=456,
            args={"test_arg": "test_value"},
            failure_message="test failure message",
            status_message="test status message",
            status_args=["Clusters[0].Status", "Clusters[0].StatusDetails", "Clusters[0].ClusterName"],
        )
        assert mock_waiter.wait.call_count == 3
        mock_sleep.assert_called_with(123)

    @mock.patch.object(_LazyStatusFormatter, "__str__")
    def test_status_formatting_not_done_if_higher_log_level(self, status_format_mock: mock.MagicMock):
        status_format_mock.return_value = "test_status"
        mock_waiter = mock.MagicMock()
        error = WaiterError(
            name="test_waiter",
            reason="test_reason",
            last_response=generate_response("Pending"),
        )
        logger = logging.getLogger(wait.__module__)
        level = logger.getEffectiveLevel()
        logger.setLevel(logging.WARNING)
        try:
            mock_waiter.wait.side_effect = [error, error, True]

            wait(
                waiter=mock_waiter,
                waiter_delay=0,
                waiter_max_attempts=456,
                args={"test_arg": "test_value"},
                failure_message="test failure message",
                status_message="test status message",
                status_args=["Status.State"],
            )
        finally:
            logger.setLevel(level)
        status_format_mock.assert_not_called()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Sequence
from typing import TYPE_CHECKING, Any

from airflow.providers.amazon.aws.hooks.kinesis_analytics import KinesisAnalyticsV2Hook
from airflow.providers.amazon.aws.sensors.base_aws import AwsBaseSensor
from airflow.providers.amazon.aws.triggers.kinesis_analytics import (
    KinesisAnalyticsV2ApplicationOperationCompleteTrigger,
)
from airflow.providers.amazon.aws.utils.mixins import aws_template_fields
from airflow.providers.common.compat.sdk import AirflowException, conf

if TYPE_CHECKING:
    from airflow.sdk import Context


class KinesisAnalyticsV2BaseSensor(AwsBaseSensor[KinesisAnalyticsV2Hook]):
    """
     General sensor behaviour for AWS Managed Service for Apache Flink.

    Subclasses must set the following fields:
        - ``INTERMEDIATE_STATES``
        - ``FAILURE_STATES``
        - ``SUCCESS_STATES``
        - ``FAILURE_MESSAGE``
        - ``SUCCESS_MESSAGE``

    :param application_name: Application name.
    :param deferrable: If True, the sensor will operate in deferrable mode. This mode requires aiobotocore
        module to be installed.
        (default: False, but can be overridden in config file by setting default_deferrable to True)

    """

    aws_hook_class = KinesisAnalyticsV2Hook
    ui_color = "#66c3ff"

    INTERMEDIATE_STATES: tuple[str, ...] = ()
    FAILURE_STATES: tuple[str, ...] = ()
    SUCCESS_STATES: tuple[str, ...] = ()
    FAILURE_MESSAGE = ""
    SUCCESS_MESSAGE = ""

    def __init__(
        self,
        application_name: str,
        deferrable: bool = conf.getboolean("operators", "default_deferrable", fallback=False),
        **kwargs: Any,
    ):
        super().__init__(**kwargs)
        self.application_name = application_name
        self.deferrable = deferrable

    def poke(self, context: Context, **kwargs) -> bool:
        status = self.hook.conn.describe_application(ApplicationName=self.application_name)[
            "ApplicationDetail"
        ]["ApplicationStatus"]

        self.log.info(
            "Poking for AWS Managed Service for Apache Flink application: %s status: %s",
            self.application_name,
            status,
        )

        if status in self.FAILURE_STATES:
            raise AirflowException(self.FAILURE_MESSAGE)

        if status in self.SUCCESS_STATES:
            self.log.info(
                "%s `%s`.",
                self.SUCCESS_MESSAGE,
                self.application_name,
            )
            return True

        return False


class KinesisAnalyticsV2StartApplicationCompletedSensor(KinesisAnalyticsV2BaseSensor):
    """
    Waits for AWS Managed Service for Apache Flink application to start.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:KinesisAnalyticsV2StartApplicationCompletedSensor`

    :param application_name: Application name.

    :param deferrable: If True, the sensor will operate in deferrable mode. This mode requires aiobotocore
        module to be installed.
        (default: False, but can be overridden in config file by setting default_deferrable to True)
    :param poke_interval: Polling period in seconds to check for the status of the job. (default: 120)
    :param max_retries: Number of times before returning the current state. (default: 75)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html

    """

    INTERMEDIATE_STATES: tuple[str, ...] = KinesisAnalyticsV2Hook.APPLICATION_START_INTERMEDIATE_STATES
    FAILURE_STATES: tuple[str, ...] = KinesisAnalyticsV2Hook.APPLICATION_START_FAILURE_STATES
    SUCCESS_STATES: tuple[str, ...] = KinesisAnalyticsV2Hook.APPLICATION_START_SUCCESS_STATES

    FAILURE_MESSAGE = "AWS Managed Service for Apache Flink application start failed."
    SUCCESS_MESSAGE = "AWS Managed Service for Apache Flink application started successfully"

    template_fields: Sequence[str] = aws_template_fields("application_name")

    def __init__(
        self,
        *,
        application_name: str,
        max_retries: int = 75,
        poke_interval: int = 120,
        **kwargs: Any,
    ) -> None:
        super().__init__(application_name=application_name, **kwargs)
        self.application_name = application_name
        self.max_retries = max_retries
        self.poke_interval = poke_interval

    def execute(self, context: Context) -> Any:
        if self.deferrable:
            self.defer(
                trigger=KinesisAnalyticsV2ApplicationOperationCompleteTrigger(
                    application_name=self.application_name,
                    waiter_name="application_start_complete",
                    aws_conn_id=self.aws_conn_id,
                    waiter_delay=int(self.poke_interval),
                    waiter_max_attempts=self.max_retries,
                    region_name=self.region_name,
                    verify=self.verify,
                    botocore_config=self.botocore_config,
                ),
                method_name="poke",
            )
        else:
            super().execute(context=context)


class KinesisAnalyticsV2StopApplicationCompletedSensor(KinesisAnalyticsV2BaseSensor):
    """
    Waits for AWS Managed Service for Apache Flink application to stop.

    .. seealso::
        For more information on how to use this sensor, take a look at the guide:
        :ref:`howto/sensor:KinesisAnalyticsV2StopApplicationCompletedSensor`

    :param application_name: Application name.

    :param deferrable: If True, the sensor will operate in deferrable mode. This mode requires aiobotocore
        module to be installed.
        (default: False, but can be overridden in config file by setting default_deferrable to True)
    :param poke_interval: Polling period in seconds to check for the status of the job. (default: 120)
    :param max_retries: Number of times before returning the current state. (default: 75)
    :param aws_conn_id: The Airflow connection used for AWS credentials.
        If this is ``None`` or empty then the default boto3 behaviour is used. If
        running Airflow in a distributed manner and aws_conn_id is None or
        empty, then default boto3 configuration would be used (and must be
        maintained on each worker node).
    :param region_name: AWS region_name. If not specified then the default boto3 behaviour is used.
    :param verify: Whether to verify SSL certificates. See:
        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/core/session.html
    :param botocore_config: Configuration dictionary (key-values) for botocore client. See:
        https://botocore.amazonaws.com/v1/documentation/api/latest/reference/config.html

    """

    INTERMEDIATE_STATES: tuple[str, ...] = KinesisAnalyticsV2Hook.APPLICATION_STOP_INTERMEDIATE_STATES
    FAILURE_STATES: tuple[str, ...] = KinesisAnalyticsV2Hook.APPLICATION_STOP_FAILURE_STATES
    SUCCESS_STATES: tuple[str, ...] = KinesisAnalyticsV2Hook.APPLICATION_STOP_SUCCESS_STATES

    FAILURE_MESSAGE = "AWS Managed Service for Apache Flink application stop failed."
    SUCCESS_MESSAGE = "AWS Managed Service for Apache Flink application stopped successfully"

    template_fields: Sequence[str] = aws_template_fields("application_name")

    def __init__(
        self,
        *,
        application_name: str,
        max_retries: int = 75,
        poke_interval: int = 120,
        **kwargs: Any,
    ) -> None:
        super().__init__(application_name=application_name, **kwargs)
        self.application_name = application_name
        self.max_retries = max_retries
        self.poke_interval = poke_interval

    def execute(self, context: Context) -> Any:
        if self.deferrable:
            self.defer(
                trigger=KinesisAnalyticsV2ApplicationOperationCompleteTrigger(
                    application_name=self.application_name,
                    waiter_name="application_stop_complete",
                    aws_conn_id=self.aws_conn_id,
                    waiter_delay=int(self.poke_interval),
                    waiter_max_attempts=self.max_retries,
                    region_name=self.region_name,
                    verify=self.verify,
                    botocore_config=self.botocore_config,
                ),
                method_name="poke",
            )
        else:
            super().execute(context=context)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import contextlib
from io import StringIO
from unittest import mock

import pytest
import time_machine

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS


class TestGetEksToken:
    @time_machine.travel("1995-02-14", tick=False)
    @pytest.mark.parametrize(
        ("args", "expected_region_name"),
        [
            [
                [
                    "airflow.providers.amazon.aws.utils.eks_get_token",
                    "--region-name",
                    "test-region",
                    "--cluster-name",
                    "test-cluster",
                    "--sts-url",
                    "https://sts.test-region.amazonaws.com/?Action=GetCallerIdentity&Version=2011-06-15",
                ],
                "test-region",
            ],
            [
                [
                    "airflow.providers.amazon.src.airflow.providers.amazon.aws.utils.eks_get_token"
                    if AIRFLOW_V_3_0_PLUS
                    else "airflow.providers.amazon.aws.utils.eks_get_token",
                    "--region-name",
                    "test-region",
                    "--cluster-name",
                    "test-cluster",
                    "--sts-url",
                    "https://sts.test-region.amazonaws.com/?Action=GetCallerIdentity&Version=2011-06-15",
                ],
                "test-region",
            ],
            [
                [
                    "airflow.providers.amazon.aws.utils.eks_get_token",
                    "--cluster-name",
                    "test-cluster",
                    "--sts-url",
                    "https://sts.us-east-1.amazonaws.com/?Action=GetCallerIdentity&Version=2011-06-15",
                ],
                None,
            ],
        ],
    )
    def test_run(self, args, expected_region_name):
        # Instead of trying to mock deep into the CLI execution context, mock the main function itself
        with mock.patch(
            "airflow.providers.amazon.aws.utils.eks_get_token.fetch_access_token_for_cluster"
        ) as mock_fetch_token:
            mock_fetch_token.return_value = "k8s-aws-v1.aHR0cDovL2V4YW1wbGUuY29t"

            with mock.patch("sys.argv", args), contextlib.redirect_stdout(StringIO()) as temp_stdout:
                # Import and directly call the main function rather than using runpy
                from airflow.providers.amazon.aws.utils.eks_get_token import main

                main()

                output = temp_stdout.getvalue()
                token = "token: k8s-aws-v1.aHR0cDovL2V4YW1wbGUuY29t"
                expected_token = output.split(",")[1].strip()
                expected_expiration_timestamp = output.split(",")[0].split(":")[1].strip()
                assert expected_token == token
                assert expected_expiration_timestamp.startswith("1995-02-")

                # Extract the sts-url from args
                sts_url = None
                if "--sts-url" in args:
                    sts_url_idx = args.index("--sts-url") + 1
                    if sts_url_idx < len(args):
                        sts_url = args[sts_url_idx]

                # Verify fetch_access_token_for_cluster was called with correct parameters
                mock_fetch_token.assert_called_once_with(
                    "test-cluster", sts_url, region_name=expected_region_name
                )

    @mock.patch("airflow.providers.amazon.aws.utils.eks_get_token.RequestSigner")
    @mock.patch("boto3.Session")
    def test_fetch_access_token_for_cluster(self, mock_session, mock_signer):
        """Test the standalone fetch_access_token_for_cluster function."""
        from airflow.providers.amazon.aws.utils.eks_get_token import fetch_access_token_for_cluster

        # Mock the session and client
        mock_session_instance = mock_session.return_value
        mock_eks_client = mock_session_instance.client.return_value
        mock_session_instance.region_name = "us-east-1"

        # Mock the RequestSigner
        mock_signer_instance = mock_signer.return_value
        mock_signer_instance.generate_presigned_url.return_value = "http://example.com"

        result = fetch_access_token_for_cluster(
            eks_cluster_name="test-cluster",
            sts_url="https://sts.us-east-1.amazonaws.com/?Action=GetCallerIdentity&Version=2011-06-15",
            region_name="us-east-1",
        )

        # Verify session creation
        mock_session.assert_called_once_with(region_name="us-east-1")
        mock_session_instance.client.assert_called_once_with("eks")

        # Verify RequestSigner was called correctly
        mock_signer.assert_called_once_with(
            service_id=mock_eks_client.meta.service_model.service_id,
            region_name="us-east-1",
            signing_name="sts",
            signature_version="v4",
            credentials=mock_session_instance.get_credentials.return_value,
            event_emitter=mock_session_instance.events,
        )

        # Verify presigned URL generation
        mock_signer_instance.generate_presigned_url.assert_called_once_with(
            request_dict={
                "method": "GET",
                "url": "https://sts.us-east-1.amazonaws.com/?Action=GetCallerIdentity&Version=2011-06-15",
                "body": {},
                "headers": {"x-k8s-aws-id": "test-cluster"},
                "context": {},
            },
            region_name="us-east-1",
            expires_in=60,
            operation_name="",
        )

        # Verify the token format
        assert result == "k8s-aws-v1.aHR0cDovL2V4YW1wbGUuY29t"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from unittest import mock
from unittest.mock import AsyncMock

import pytest

from airflow.providers.amazon.aws.hooks.bedrock import BedrockAgentHook, BedrockHook
from airflow.providers.amazon.aws.triggers.bedrock import (
    BedrockBatchInferenceCompletedTrigger,
    BedrockBatchInferenceScheduledTrigger,
    BedrockCustomizeModelCompletedTrigger,
    BedrockIngestionJobTrigger,
    BedrockKnowledgeBaseActiveTrigger,
    BedrockProvisionModelThroughputCompletedTrigger,
)
from airflow.triggers.base import TriggerEvent

from unit.amazon.aws.utils.test_waiter import assert_expected_waiter_type

BASE_TRIGGER_CLASSPATH = "airflow.providers.amazon.aws.triggers.bedrock."


class TestBaseBedrockTrigger:
    EXPECTED_WAITER_NAME: str | None = None

    def test_setup(self):
        # Ensure that all subclasses have an expected waiter name set.
        if self.__class__.__name__ != "TestBaseBedrockTrigger":
            assert isinstance(self.EXPECTED_WAITER_NAME, str)


class TestBedrockCustomizeModelCompletedTrigger(TestBaseBedrockTrigger):
    EXPECTED_WAITER_NAME = "model_customization_job_complete"
    JOB_NAME = "test_job"

    def test_serialization(self):
        """Assert that arguments and classpath are correctly serialized."""
        trigger = BedrockCustomizeModelCompletedTrigger(job_name=self.JOB_NAME)
        classpath, kwargs = trigger.serialize()
        assert classpath == BASE_TRIGGER_CLASSPATH + "BedrockCustomizeModelCompletedTrigger"
        assert kwargs.get("job_name") == self.JOB_NAME

    @pytest.mark.asyncio
    @mock.patch.object(BedrockHook, "get_waiter")
    @mock.patch.object(BedrockHook, "get_async_conn")
    async def test_run_success(self, mock_async_conn, mock_get_waiter):
        mock_async_conn.__aenter__.return_value = mock.MagicMock()
        mock_get_waiter().wait = AsyncMock()
        trigger = BedrockCustomizeModelCompletedTrigger(job_name=self.JOB_NAME)

        generator = trigger.run()
        response = await generator.asend(None)

        assert response == TriggerEvent({"status": "success", "job_name": self.JOB_NAME})
        assert_expected_waiter_type(mock_get_waiter, self.EXPECTED_WAITER_NAME)
        mock_get_waiter().wait.assert_called_once()


class TestBedrockProvisionModelThroughputCompletedTrigger(TestBaseBedrockTrigger):
    EXPECTED_WAITER_NAME = "provisioned_model_throughput_complete"
    PROVISIONED_MODEL_ID = "provisioned_model_id"

    def test_serialization(self):
        """Assert that arguments and classpath are correctly serialized."""
        trigger = BedrockProvisionModelThroughputCompletedTrigger(
            provisioned_model_id=self.PROVISIONED_MODEL_ID
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == BASE_TRIGGER_CLASSPATH + "BedrockProvisionModelThroughputCompletedTrigger"
        assert kwargs.get("provisioned_model_id") == self.PROVISIONED_MODEL_ID

    @pytest.mark.asyncio
    @mock.patch.object(BedrockHook, "get_waiter")
    @mock.patch.object(BedrockHook, "get_async_conn")
    async def test_run_success(self, mock_async_conn, mock_get_waiter):
        mock_async_conn.__aenter__.return_value = mock.MagicMock()
        mock_get_waiter().wait = AsyncMock()
        trigger = BedrockProvisionModelThroughputCompletedTrigger(
            provisioned_model_id=self.PROVISIONED_MODEL_ID
        )

        generator = trigger.run()
        response = await generator.asend(None)

        assert response == TriggerEvent(
            {"status": "success", "provisioned_model_id": self.PROVISIONED_MODEL_ID}
        )
        assert_expected_waiter_type(mock_get_waiter, self.EXPECTED_WAITER_NAME)
        mock_get_waiter().wait.assert_called_once()


class TestBedrockKnowledgeBaseActiveTrigger(TestBaseBedrockTrigger):
    EXPECTED_WAITER_NAME = "knowledge_base_active"
    KNOWLEDGE_BASE_NAME = "test_kb"

    def test_serialization(self):
        """Assert that arguments and classpath are correctly serialized."""
        trigger = BedrockKnowledgeBaseActiveTrigger(knowledge_base_id=self.KNOWLEDGE_BASE_NAME)
        classpath, kwargs = trigger.serialize()
        assert classpath == BASE_TRIGGER_CLASSPATH + "BedrockKnowledgeBaseActiveTrigger"
        assert kwargs.get("knowledge_base_id") == self.KNOWLEDGE_BASE_NAME

    @pytest.mark.asyncio
    @mock.patch.object(BedrockAgentHook, "get_waiter")
    @mock.patch.object(BedrockAgentHook, "get_async_conn")
    async def test_run_success(self, mock_async_conn, mock_get_waiter):
        mock_async_conn.__aenter__.return_value = mock.MagicMock()
        mock_get_waiter().wait = AsyncMock()
        trigger = BedrockKnowledgeBaseActiveTrigger(knowledge_base_id=self.KNOWLEDGE_BASE_NAME)

        generator = trigger.run()
        response = await generator.asend(None)

        assert response == TriggerEvent({"status": "success", "knowledge_base_id": self.KNOWLEDGE_BASE_NAME})
        assert_expected_waiter_type(mock_get_waiter, self.EXPECTED_WAITER_NAME)
        mock_get_waiter().wait.assert_called_once()


class TestBedrockIngestionJobTrigger(TestBaseBedrockTrigger):
    EXPECTED_WAITER_NAME = "ingestion_job_complete"

    KNOWLEDGE_BASE_ID = "test_kb"
    DATA_SOURCE_ID = "test_ds"
    INGESTION_JOB_ID = "test_ingestion_job"

    def test_serialization(self):
        """Assert that arguments and classpath are correctly serialized."""
        trigger = BedrockIngestionJobTrigger(
            knowledge_base_id=self.KNOWLEDGE_BASE_ID,
            data_source_id=self.DATA_SOURCE_ID,
            ingestion_job_id=self.INGESTION_JOB_ID,
        )
        classpath, kwargs = trigger.serialize()
        assert classpath == BASE_TRIGGER_CLASSPATH + "BedrockIngestionJobTrigger"
        assert kwargs.get("knowledge_base_id") == self.KNOWLEDGE_BASE_ID
        assert kwargs.get("data_source_id") == self.DATA_SOURCE_ID
        assert kwargs.get("ingestion_job_id") == self.INGESTION_JOB_ID

    @pytest.mark.asyncio
    @mock.patch.object(BedrockAgentHook, "get_waiter")
    @mock.patch.object(BedrockAgentHook, "get_async_conn")
    async def test_run_success(self, mock_async_conn, mock_get_waiter):
        mock_async_conn.__aenter__.return_value = mock.MagicMock()
        mock_get_waiter().wait = AsyncMock()
        trigger = BedrockIngestionJobTrigger(
            knowledge_base_id=self.KNOWLEDGE_BASE_ID,
            data_source_id=self.DATA_SOURCE_ID,
            ingestion_job_id=self.INGESTION_JOB_ID,
        )

        generator = trigger.run()
        response = await generator.asend(None)

        assert_expected_waiter_type(mock_get_waiter, self.EXPECTED_WAITER_NAME)
        assert response == TriggerEvent({"status": "success", "ingestion_job_id": self.INGESTION_JOB_ID})
        mock_get_waiter().wait.assert_called_once()


class TestBedrockBatchInferenceCompletedTrigger(TestBaseBedrockTrigger):
    EXPECTED_WAITER_NAME = "batch_inference_complete"

    JOB_ARN = "job_arn"

    def test_serialization(self):
        """Assert that arguments and classpath are correctly serialized."""
        trigger = BedrockBatchInferenceCompletedTrigger(job_arn=self.JOB_ARN)

        classpath, kwargs = trigger.serialize()

        assert classpath == BASE_TRIGGER_CLASSPATH + "BedrockBatchInferenceCompletedTrigger"
        assert kwargs.get("job_arn") == self.JOB_ARN

    @pytest.mark.asyncio
    @mock.patch.object(BedrockHook, "get_waiter")
    @mock.patch.object(BedrockHook, "get_async_conn")
    async def test_run_success(self, mock_async_conn, mock_get_waiter):
        mock_async_conn.__aenter__.return_value = mock.MagicMock()
        mock_get_waiter().wait = AsyncMock()
        trigger = BedrockBatchInferenceCompletedTrigger(job_arn=self.JOB_ARN)

        generator = trigger.run()
        response = await generator.asend(None)

        assert_expected_waiter_type(mock_get_waiter, self.EXPECTED_WAITER_NAME)
        assert response == TriggerEvent({"status": "success", "job_arn": self.JOB_ARN})
        mock_get_waiter().wait.assert_called_once()


class TestBedrockBatchInferenceScheduledTrigger(TestBaseBedrockTrigger):
    EXPECTED_WAITER_NAME = "batch_inference_scheduled"

    JOB_ARN = "job_arn"

    def test_serialization(self):
        """Assert that arguments and classpath are correctly serialized."""
        trigger = BedrockBatchInferenceScheduledTrigger(job_arn=self.JOB_ARN)

        classpath, kwargs = trigger.serialize()

        assert classpath == BASE_TRIGGER_CLASSPATH + "BedrockBatchInferenceScheduledTrigger"
        assert kwargs.get("job_arn") == self.JOB_ARN

    @pytest.mark.asyncio
    @mock.patch.object(BedrockHook, "get_waiter")
    @mock.patch.object(BedrockHook, "get_async_conn")
    async def test_run_success(self, mock_async_conn, mock_get_waiter):
        mock_async_conn.__aenter__.return_value = mock.MagicMock()
        mock_get_waiter().wait = AsyncMock()
        trigger = BedrockBatchInferenceScheduledTrigger(job_arn=self.JOB_ARN)

        generator = trigger.run()
        response = await generator.asend(None)

        assert_expected_waiter_type(mock_get_waiter, self.EXPECTED_WAITER_NAME)
        assert response == TriggerEvent({"status": "success", "job_arn": self.JOB_ARN})
        mock_get_waiter().wait.assert_called_once()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from airflow.providers.amazon.aws.auth_manager.constants import (
    CONF_AVP_POLICY_STORE_ID_KEY,
    CONF_CONN_ID_KEY,
    CONF_REGION_NAME_KEY,
    CONF_SAML_METADATA_URL_KEY,
    CONF_SECTION_NAME,
)


class TestAwsAuthManagerConstants:
    def test_conf_section_name(self):
        assert CONF_SECTION_NAME == "aws_auth_manager"

    def test_conf_conn_id_key(self):
        assert CONF_CONN_ID_KEY == "conn_id"

    def test_conf_region_name_key(self):
        assert CONF_REGION_NAME_KEY == "region_name"

    def test_conf_saml_metadata_url_key(self):
        assert CONF_SAML_METADATA_URL_KEY == "saml_metadata_url"

    def test_conf_avp_policy_store_id_key(self):
        assert CONF_AVP_POLICY_STORE_ID_KEY == "avp_policy_store_id"



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from datetime import datetime

import boto3

from airflow.providers.amazon.aws.hooks.s3 import S3Hook
from airflow.providers.amazon.aws.operators.athena import AthenaOperator
from airflow.providers.amazon.aws.operators.s3 import (
    S3CreateBucketOperator,
    S3CreateObjectOperator,
    S3DeleteBucketOperator,
)
from airflow.providers.amazon.aws.sensors.athena import AthenaSensor

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import SystemTestContextBuilder

sys_test_context_task = SystemTestContextBuilder().build()

DAG_ID = "example_athena"

SAMPLE_DATA = """"Alice",20
    "Bob",25
    "Charlie",30
    """
SAMPLE_FILENAME = "airflow_sample.csv"


@task
def await_bucket(bucket_name):
    # Avoid a race condition after creating the S3 Bucket.
    client = boto3.client("s3")
    waiter = client.get_waiter("bucket_exists")
    waiter.wait(Bucket=bucket_name)


@task
def read_results_from_s3(bucket_name, query_execution_id):
    s3_hook = S3Hook()
    file_obj = s3_hook.get_conn().get_object(Bucket=bucket_name, Key=f"{query_execution_id}.csv")
    file_content = file_obj["Body"].read().decode("utf-8")
    print(file_content)


with DAG(
    dag_id=DAG_ID,
    schedule="@once",
    start_date=datetime(2021, 1, 1),
    catchup=False,
) as dag:
    test_context = sys_test_context_task()
    env_id = test_context["ENV_ID"]

    s3_bucket = f"{env_id}-athena-bucket"
    athena_table = f"{env_id}_test_table"
    athena_database = f"{env_id}_default"

    query_create_database = f"CREATE DATABASE IF NOT EXISTS {athena_database}"
    query_create_table = f"""CREATE EXTERNAL TABLE IF NOT EXISTS {athena_database}.{athena_table}
        ( `name` string, `age` int )
        ROW FORMAT SERDE "org.apache.hadoop.hive.serde2.lazy.LazySimpleSerDe"
        WITH SERDEPROPERTIES ( "serialization.format" = ",", "field.delim" = "," )
        LOCATION "s3://{s3_bucket}//{athena_table}"
        TBLPROPERTIES ("has_encrypted_data"="false")
        """
    query_read_table = f"SELECT * from {athena_database}.{athena_table}"
    query_drop_table = f"DROP TABLE IF EXISTS {athena_database}.{athena_table}"
    query_drop_database = f"DROP DATABASE IF EXISTS {athena_database}"

    create_s3_bucket = S3CreateBucketOperator(task_id="create_s3_bucket", bucket_name=s3_bucket)

    upload_sample_data = S3CreateObjectOperator(
        task_id="upload_sample_data",
        s3_bucket=s3_bucket,
        s3_key=f"{athena_table}/{SAMPLE_FILENAME}",
        data=SAMPLE_DATA,
        replace=True,
    )

    create_database = AthenaOperator(
        task_id="create_database",
        query=query_create_database,
        database=athena_database,
        output_location=f"s3://{s3_bucket}/",
        sleep_time=1,
    )

    create_table = AthenaOperator(
        task_id="create_table",
        query=query_create_table,
        database=athena_database,
        output_location=f"s3://{s3_bucket}/",
        sleep_time=1,
    )

    # [START howto_operator_athena]
    read_table = AthenaOperator(
        task_id="read_table",
        query=query_read_table,
        database=athena_database,
        output_location=f"s3://{s3_bucket}/",
    )
    # [END howto_operator_athena]
    read_table.sleep_time = 1

    # [START howto_sensor_athena]
    await_query = AthenaSensor(
        task_id="await_query",
        query_execution_id=read_table.output,
    )
    # [END howto_sensor_athena]

    drop_table = AthenaOperator(
        task_id="drop_table",
        query=query_drop_table,
        database=athena_database,
        output_location=f"s3://{s3_bucket}/",
        trigger_rule=TriggerRule.ALL_DONE,
        sleep_time=1,
    )

    drop_database = AthenaOperator(
        task_id="drop_database",
        query=query_drop_database,
        database=athena_database,
        output_location=f"s3://{s3_bucket}/",
        trigger_rule=TriggerRule.ALL_DONE,
        sleep_time=1,
    )

    delete_s3_bucket = S3DeleteBucketOperator(
        task_id="delete_s3_bucket",
        bucket_name=s3_bucket,
        force_delete=True,
        trigger_rule=TriggerRule.ALL_DONE,
    )

    chain(
        # TEST SETUP
        test_context,
        create_s3_bucket,
        await_bucket(s3_bucket),
        upload_sample_data,
        create_database,
        # TEST BODY
        create_table,
        read_table,
        await_query,
        read_results_from_s3(s3_bucket, read_table.output),
        # TEST TEARDOWN
        drop_table,
        drop_database,
        delete_s3_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

import boto3
import pendulum

from airflow.providers.amazon.aws.operators.glue_databrew import (
    GlueDataBrewStartJobOperator,
)
from airflow.providers.amazon.aws.operators.s3 import (
    S3CreateBucketOperator,
    S3CreateObjectOperator,
    S3DeleteBucketOperator,
)

from tests_common.test_utils.version_compat import AIRFLOW_V_3_0_PLUS

if AIRFLOW_V_3_0_PLUS:
    from airflow.sdk import DAG, chain, task
else:
    # Airflow 2 path
    from airflow.decorators import task  # type: ignore[attr-defined,no-redef]
    from airflow.models.baseoperator import chain  # type: ignore[attr-defined,no-redef]
    from airflow.models.dag import DAG  # type: ignore[attr-defined,no-redef,assignment]

try:
    from airflow.sdk import TriggerRule
except ImportError:
    # Compatibility for Airflow < 3.1
    from airflow.utils.trigger_rule import TriggerRule  # type: ignore[no-redef,attr-defined]

from system.amazon.aws.utils import SystemTestContextBuilder

DAG_ID = "example_glue_databrew"

# Externally fetched variables:
ROLE_ARN_KEY = "ROLE_ARN"

sys_test_context_task = SystemTestContextBuilder().add_variable(ROLE_ARN_KEY).build()

EXAMPLE_JSON = "{}"


@task
def create_dataset(dataset_name: str, bucket_name: str, object_key: str):
    client = boto3.client("databrew")
    client.create_dataset(
        Name=dataset_name,
        Format="JSON",
        FormatOptions={
            "Json": {"MultiLine": False},
        },
        Input={
            "S3InputDefinition": {
                "Bucket": bucket_name,
                "Key": object_key,
            },
        },
    )


@task
def create_job(
    dataset_name: str, job_name: str, bucket_output_name: str, object_output_key: str, role_arn: str
):
    client = boto3.client("databrew")
    client.create_profile_job(
        DatasetName=dataset_name,
        Name=job_name,
        LogSubscription="ENABLE",
        OutputLocation={
            "Bucket": bucket_output_name,
            "Key": object_output_key,
        },
        RoleArn=role_arn,
    )


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_dataset(dataset_name: str):
    client = boto3.client("databrew")
    client.delete_dataset(Name=dataset_name)


@task(trigger_rule=TriggerRule.ALL_DONE)
def delete_job(job_name: str):
    client = boto3.client("databrew")
    client.delete_job(Name=job_name)


with DAG(DAG_ID, schedule="@once", start_date=pendulum.datetime(2023, 1, 1, tz="UTC"), catchup=False) as dag:
    test_context = sys_test_context_task()
    env_id = test_context["ENV_ID"]
    role_arn = test_context[ROLE_ARN_KEY]

    bucket_name = f"{env_id}-bucket-databrew"
    output_bucket_name = f"{env_id}-output-bucket-databrew"
    file_name = "data.json"
    dataset_name = f"{env_id}-dataset"
    job_name = f"{env_id}-databrew-job"

    create_bucket = S3CreateBucketOperator(
        task_id="create_bucket",
        bucket_name=bucket_name,
    )

    create_output_bucket = S3CreateBucketOperator(
        task_id="create_output_bucket",
        bucket_name=output_bucket_name,
    )

    upload_file = S3CreateObjectOperator(
        task_id="upload_file",
        s3_bucket=bucket_name,
        s3_key=file_name,
        data=EXAMPLE_JSON,
        replace=True,
    )

    # [START howto_operator_glue_databrew_start]
    start_job = GlueDataBrewStartJobOperator(task_id="startjob", job_name=job_name, waiter_delay=15)
    # [END howto_operator_glue_databrew_start]

    delete_bucket = S3DeleteBucketOperator(
        task_id="delete_bucket",
        trigger_rule=TriggerRule.ALL_DONE,
        bucket_name=bucket_name,
        force_delete=True,
    )

    delete_output_bucket = S3DeleteBucketOperator(
        task_id="delete_output_bucket",
        trigger_rule=TriggerRule.ALL_DONE,
        bucket_name=output_bucket_name,
        force_delete=True,
    )

    chain(
        # TEST SETUP
        test_context,
        create_bucket,
        create_output_bucket,
        upload_file,
        create_dataset(dataset_name, bucket_name, file_name),
        create_job(dataset_name, job_name, output_bucket_name, "output.json", role_arn),
        # TEST BODY
        start_job,
        # TEST TEARDOWN
        delete_job(job_name),
        delete_dataset(dataset_name),
        delete_bucket,
        delete_output_bucket,
    )

    from tests_common.test_utils.watcher import watcher

    # This test needs watcher in order to properly mark success/failure
    # when "tearDown" task with trigger rule is part of the DAG
    list(dag.tasks) >> watcher()

from tests_common.test_utils.system_tests import get_test_run  # noqa: E402

# Needed to run the example DAG with pytest (see: tests/system/README.md#run_via_pytest)
test_run = get_test_run(dag)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from urllib.parse import quote_plus

from airflow.providers.amazon.aws.links.base_aws import BASE_AWS_CONSOLE_LINK, BaseAwsLink


class StateMachineDetailsLink(BaseAwsLink):
    """Helper class for constructing link to State Machine details page."""

    name = "State Machine Details"
    key = "_state_machine_details"
    format_str = (
        BASE_AWS_CONSOLE_LINK + "/states/home?region={region_name}#/statemachines/view/{state_machine_arn}"
    )

    def format_link(self, *, state_machine_arn: str | None = None, **kwargs) -> str:
        if not state_machine_arn:
            return ""
        return super().format_link(state_machine_arn=quote_plus(state_machine_arn), **kwargs)


class StateMachineExecutionsDetailsLink(BaseAwsLink):
    """Helper class for constructing link to State Machine Execution details page."""

    name = "State Machine Executions Details"
    key = "_state_machine_executions_details"
    format_str = (
        BASE_AWS_CONSOLE_LINK + "/states/home?region={region_name}#/v2/executions/details/{execution_arn}"
    )

    def format_link(self, *, execution_arn: str | None = None, **kwargs) -> str:
        if not execution_arn:
            return ""
        return super().format_link(execution_arn=quote_plus(execution_arn), **kwargs)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
from __future__ import annotations

from collections.abc import Iterable, Sequence
from functools import cached_property
from typing import Any

from airflow.providers.amazon.aws.hooks.ses import SesHook
from airflow.providers.amazon.version_compat import AIRFLOW_V_3_1_PLUS
from airflow.providers.common.compat.notifier import BaseNotifier
from airflow.utils.helpers import prune_dict


class SesNotifier(BaseNotifier):
    """
    Amazon Simple Email Service (SES) Notifier.

    :param mail_from: Email address to set as email's from
    :param to: List of email addresses to set as email's to
    :param subject: Email's subject
    :param html_content: Content of email in HTML format
    :param files: List of paths of files to be attached
    :param cc: List of email addresses to set as email's CC
    :param bcc: List of email addresses to set as email's BCC
    :param mime_subtype: Can be used to specify the subtype of the message. Default = mixed
    :param mime_charset: Email's charset. Default = UTF-8.
    :param return_path: The email address to which replies will be sent. By default, replies
        are sent to the original sender's email address.
    :param reply_to: The email address to which message bounces and complaints should be sent.
        "Return-Path" is sometimes called "envelope from", "envelope sender", or "MAIL FROM".
    :param custom_headers: Additional headers to add to the MIME message.
        No validations are run on these values, and they should be able to be encoded.
    """

    template_fields: Sequence[str] = (
        "aws_conn_id",
        "region_name",
        "mail_from",
        "to",
        "subject",
        "html_content",
        "files",
        "cc",
        "bcc",
        "mime_subtype",
        "mime_charset",
        "reply_to",
        "return_path",
        "custom_headers",
    )

    def __init__(
        self,
        *,
        aws_conn_id: str | None = SesHook.default_conn_name,
        region_name: str | None = None,
        mail_from: str,
        to: str | Iterable[str],
        subject: str,
        html_content: str,
        files: list[str] | None = None,
        cc: str | Iterable[str] | None = None,
        bcc: str | Iterable[str] | None = None,
        mime_subtype: str = "mixed",
        mime_charset: str = "utf-8",
        reply_to: str | None = None,
        return_path: str | None = None,
        custom_headers: dict[str, Any] | None = None,
        **kwargs,
    ):
        if AIRFLOW_V_3_1_PLUS:
            #  Support for passing context was added in 3.1.0
            super().__init__(**kwargs)
        else:
            super().__init__()
        self.aws_conn_id = aws_conn_id
        self.region_name = region_name

        self.mail_from = mail_from
        self.to = to
        self.subject = subject
        self.html_content = html_content
        self.files = files
        self.cc = cc
        self.bcc = bcc
        self.mime_subtype = mime_subtype
        self.mime_charset = mime_charset
        self.reply_to = reply_to
        self.return_path = return_path
        self.custom_headers = custom_headers

    def _build_send_kwargs(self):
        return prune_dict(
            {
                "mail_from": self.mail_from,
                "to": self.to,
                "subject": self.subject,
                "html_content": self.html_content,
                "files": self.files,
                "cc": self.cc,
                "bcc": self.bcc,
                "mime_subtype": self.mime_subtype,
                "mime_charset": self.mime_charset,
                "reply_to": self.reply_to,
                "return_path": self.return_path,
                "custom_headers": self.custom_headers,
            }
        )

    @cached_property
    def hook(self) -> SesHook:
        """Amazon Simple Email Service (SES) Hook (cached)."""
        return SesHook(aws_conn_id=self.aws_conn_id, region_name=self.region_name)

    def notify(self, context):
        """Send email using Amazon Simple Email Service (SES)."""
        self.hook.send_email(**self._build_send_kwargs())

    async def async_notify(self, context):
        """Send email using Amazon Simple Email Service (SES) (async)."""
        await self.hook.asend_email(**self._build_send_kwargs())


send_ses_notification = SesNotifier
